
# -*- coding: utf-8 -*-


# PyMeeus: Python module implementing astronomical algorithms.
# Copyright (C) 2021  Dagoberto Salazar
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import sin, cos, asin, atan2, tan, sqrt
from pymeeus.Angle import Angle
from pymeeus.Epoch import Epoch, JDE2000
from pymeeus.Sun import Sun
from pymeeus.Coordinates import (
    nutation_longitude, true_obliquity, ecliptical2equatorial
)

"""
.. module:: Moon
   :synopsis: Class to model the Moon
   :license: GNU Lesser General Public License v3 (LGPLv3)

.. moduleauthor:: Dagoberto Salazar
"""


PERIODIC_TERMS_LR_TABLE = [
    [0,  0,  1,  0, 6288774.0, -20905355.0],
    [2,  0, -1,  0, 1274027.0,  -3699111.0],
    [2,  0,  0,  0,  658314.0,  -2955968.0],
    [0,  0,  2,  0,  213618.0,   -569925.0],
    [0,  1,  0,  0, -185116.0,     48888.0],
    [0,  0,  0,  2, -114332.0,     -3149.0],
    [2,  0, -2,  0,   58793.0,    246158.0],
    [2, -1, -1,  0,   57066.0,   -152138.0],
    [2,  0,  1,  0,   53322.0,   -170733.0],
    [2, -1,  0,  0,   45758.0,   -204586.0],
    [0,  1, -1,  0,  -40923.0,   -129620.0],
    [1,  0,  0,  0,  -34720.0,    108743.0],
    [0,  1,  1,  0,  -30383.0,    104755.0],
    [2,  0,  0, -2,   15327.0,     10321.0],
    [0,  0,  1,  2,  -12528.0,         0.0],
    [0,  0,  1, -2,   10980.0,     79661.0],
    [4,  0, -1,  0,   10675.0,    -34782.0],
    [0,  0,  3,  0,   10034.0,    -23210.0],
    [4,  0, -2,  0,    8548.0,    -21636.0],
    [2,  1, -1,  0,   -7888.0,     24208.0],
    [2,  1,  0,  0,   -6766.0,     30824.0],
    [1,  0, -1,  0,   -5163.0,     -8379.0],
    [1,  1,  0,  0,    4987.0,    -16675.0],
    [2, -1,  1,  0,    4036.0,    -12831.0],
    [2,  0,  2,  0,    3994.0,    -10445.0],
    [4,  0,  0,  0,    3861.0,    -11650.0],
    [2,  0, -3,  0,    3665.0,     14403.0],
    [0,  1, -2,  0,   -2689.0,     -7003.0],
    [2,  0, -1,  2,   -2602.0,         0.0],
    [2, -1, -2,  0,    2390.0,     10056.0],
    [1,  0,  1,  0,   -2348.0,      6322.0],
    [2, -2,  0,  0,    2236.0,     -9884.0],
    [0,  1,  2,  0,   -2120.0,      5751.0],
    [0,  2,  0,  0,   -2069.0,         0.0],
    [2, -2, -1,  0,    2048.0,     -4950.0],
    [2,  0,  1, -2,   -1773.0,      4130.0],
    [2,  0,  0,  2,   -1595.0,         0.0],
    [4, -1, -1,  0,    1215.0,     -3958.0],
    [0,  0,  2,  2,   -1110.0,         0.0],
    [3,  0, -1,  0,    -892.0,      3258.0],
    [2,  1,  1,  0,    -810.0,      2616.0],
    [4, -1, -2,  0,     759.0,     -1897.0],
    [0,  2, -1,  0,    -713.0,     -2117.0],
    [2,  2, -1,  0,    -700.0,      2354.0],
    [2,  1, -2,  0,     691.0,         0.0],
    [2, -1,  0, -2,     596.0,         0.0],
    [4,  0,  1,  0,     549.0,     -1423.0],
    [0,  0,  4,  0,     537.0,     -1117.0],
    [4, -1,  0,  0,     520.0,     -1571.0],
    [1,  0, -2,  0,    -487.0,     -1739.0],
    [2,  1,  0, -2,    -399.0,         0.0],
    [0,  0,  2, -2,    -381.0,     -4421.0],
    [1,  1,  1,  0,     351.0,         0.0],
    [3,  0, -2,  0,    -340.0,         0.0],
    [4,  0, -3,  0,     330.0,         0.0],
    [2, -1,  2,  0,     327.0,         0.0],
    [0,  2,  1,  0,    -323.0,      1165.0],
    [1,  1, -1,  0,     299.0,         0.0],
    [2,  0,  3,  0,     294.0,         0.0],
    [2,  0, -1, -2,       0.0,      8752.0]
]
"""This table contains the periodic terms for the longitude (Sigmal) and
distance (Sigmar) of the Moon. Units are 0.000001 degree for Sigmal, and 0.001
kilometer for Sigmar. In Meeus' book this is Table 47.A and can be found in
pages 339-340."""

PERIODIC_TERMS_B_TABLE = [
    [0,  0,  0,  1, 5128122.0],
    [0,  0,  1,  1,  280602.0],
    [0,  0,  1, -1,  277693.0],
    [2,  0,  0, -1,  173237.0],
    [2,  0, -1,  1,   55413.0],
    [2,  0, -1, -1,   46271.0],
    [2,  0,  0,  1,   32573.0],
    [0,  0,  2,  1,   17198.0],
    [2,  0,  1, -1,    9266.0],
    [0,  0,  2, -1,    8822.0],
    [2, -1,  0, -1,    8216.0],
    [2,  0, -2, -1,    4324.0],
    [2,  0,  1,  1,    4200.0],
    [2,  1,  0, -1,   -3359.0],
    [2, -1, -1,  1,    2463.0],
    [2, -1,  0,  1,    2211.0],
    [2, -1, -1, -1,    2065.0],
    [0,  1, -1, -1,   -1870.0],
    [4,  0, -1, -1,    1828.0],
    [0,  1,  0,  1,   -1794.0],
    [0,  0,  0,  3,   -1749.0],
    [0,  1, -1,  1,   -1565.0],
    [1,  0,  0,  1,   -1491.0],
    [0,  1,  1,  1,   -1475.0],
    [0,  1,  1, -1,   -1410.0],
    [0,  1,  0, -1,   -1344.0],
    [1,  0,  0, -1,   -1335.0],
    [0,  0,  3,  1,    1107.0],
    [4,  0,  0, -1,    1021.0],
    [4,  0, -1,  1,     833.0],
    [0,  0,  1, -3,     777.0],
    [4,  0, -2,  1,     671.0],
    [2,  0,  0, -3,     607.0],
    [2,  0,  2, -1,     596.0],
    [2, -1,  1, -1,     491.0],
    [2,  0, -2,  1,    -451.0],
    [0,  0,  3, -1,     439.0],
    [2,  0,  2,  1,     422.0],
    [2,  0, -3, -1,     421.0],
    [2,  1, -1,  1,    -366.0],
    [2,  1,  0,  1,    -351.0],
    [4,  0,  0,  1,     331.0],
    [2, -1,  1,  1,     315.0],
    [2, -2,  0, -1,     302.0],
    [0,  0,  1,  3,    -283.0],
    [2,  1,  1, -1,    -229.0],
    [1,  1,  0, -1,     223.0],
    [1,  1,  0,  1,     223.0],
    [0,  1, -2, -1,    -220.0],
    [2,  1, -1, -1,    -220.0],
    [1,  0,  1,  1,    -185.0],
    [2, -1, -2, -1,     181.0],
    [0,  1,  2,  1,    -177.0],
    [4,  0, -2, -1,     176.0],
    [4, -1, -1, -1,     166.0],
    [1,  0,  1, -1,    -164.0],
    [4,  0,  1, -1,     132.0],
    [1,  0, -1, -1,    -119.0],
    [4, -1,  0, -1,     115.0],
    [2, -2,  0,  1,     107.0],
]
"""This table contains the periodic terms for the latitude of the Moon Sigmab.
Units are 0.000001 degree. In Meeus' book this is Table 47.B and can be found
in page 341."""


class Moon(object):
    """
    Class Moon models Earth's satellite.
    """

    @staticmethod
    def geocentric_ecliptical_pos(epoch):
        """This method computes the geocentric ecliptical position (longitude,
        latitude) of the Moon for a given instant, referred to the mean equinox
        of the date, as well as the Moon-Earth distance in kilometers and the
        equatorial horizontal parallax.

        :param epoch: Instant to compute the Moon's position, as an
            py:class:`Epoch` object.
        :type epoch: :py:class:`Epoch`

        :returns: Tuple containing:

            * Geocentric longitude of the center of the Moon, as an
              py:class:`Epoch` object.
            * Geocentric latitude of the center of the Moon, as an
              py:class:`Epoch` object.
            * Distance in kilometers between the centers of Earth and Moon, in
              kilometers (float)
            * Equatorial horizontal parallax of the Moon, as an
              py:class:`Epoch` object.
        :rtype: tuple
        :raises: TypeError if input value is of wrong type.

        >>> epoch = Epoch(1992, 4, 12.0)
        >>> Lambda, Beta, Delta, ppi = Moon.geocentric_ecliptical_pos(epoch)
        >>> print(round(Lambda, 6))
        133.162655
        >>> print(round(Beta, 6))
        -3.229126
        >>> print(round(Delta, 1))
        368409.7
        >>> print(round(ppi, 5))
        0.99199
        """

        # First check that input values are of correct types
        if not (isinstance(epoch, Epoch)):
            raise TypeError("Invalid input type")
        # Get the time from J2000.0 in Julian centuries
        t = (epoch - JDE2000) / 36525.0
        # Compute Moon's mean longitude, referred to mean equinox of date
        Lprime = 218.3164477 + (481267.88123421
                                + (-0.0015786
                                   + (1.0/538841.0
                                       - t/65194000.0) * t) * t) * t
        # Mean elongation of the Moon
        D = 297.8501921 + (445267.1114034
                           + (-0.0018819
                              + (1.0/545868.0 - t/113065000.0) * t) * t) * t
        # Sun's mean anomaly
        M = 357.5291092 + (35999.0502909 + (-0.0001536 + t/24490000.0) * t) * t
        # Moon's mean anomaly
        Mprime = 134.9633964 + (477198.8675055
                                + (0.0087414
                                   + (1.0/69699.9
                                      + t/14712000.0) * t) * t) * t
        # Moon's argument of latitude
        F = 93.2720950 + (483202.0175233
                          + (-0.0036539
                             + (-1.0/3526000.0 + t/863310000.0) * t) * t) * t
        # Let's compute some additional arguments
        A1 = 119.75 + 131.849 * t
        A2 = 53.09 + 479264.290 * t
        A3 = 313.45 + 481266.484 * t
        # Eccentricity of Earth's orbit around the Sun
        E = 1.0 + (-0.002516 - 0.0000074 * t) * t
        E2 = E * E
        # Reduce the angles to a [0 360] range
        Lprime = Angle(Angle.reduce_deg(Lprime)).to_positive()
        Lprimer = Lprime.rad()
        D = Angle(Angle.reduce_deg(D)).to_positive()
        Dr = D.rad()
        M = Angle(Angle.reduce_deg(M)).to_positive()
        Mr = M.rad()
        Mprime = Angle(Angle.reduce_deg(Mprime)).to_positive()
        Mprimer = Mprime.rad()
        F = Angle(Angle.reduce_deg(F)).to_positive()
        Fr = F.rad()
        A1 = Angle(Angle.reduce_deg(A1)).to_positive()
        A1r = A1.rad()
        A2 = Angle(Angle.reduce_deg(A2)).to_positive()
        A2r = A2.rad()
        A3 = Angle(Angle.reduce_deg(A3)).to_positive()
        A3r = A3.rad()
        # Let's store this results in a list, in preparation for using tables
        arguments = [Dr, Mr, Mprimer, Fr]
        # Now we use the tables of periodic terms. First for sigmal and sigmar
        sigmal = 0.0
        sigmar = 0.0
        for i, value in enumerate(PERIODIC_TERMS_LR_TABLE):
            argument = 0.0
            for j in range(4):
                if PERIODIC_TERMS_LR_TABLE[i][j]:  # Avoid multiply by zero
                    argument += PERIODIC_TERMS_LR_TABLE[i][j] * arguments[j]
            coeffl = value[4]
            coeffr = value[5]
            if abs(value[1]) == 1:
                coeffl = coeffl * E
                coeffr = coeffr * E
            elif abs(value[1]) == 2:
                coeffl = coeffl * E2
                coeffr = coeffr * E2
            sigmal += coeffl * sin(argument)
            sigmar += coeffr * cos(argument)
        # Add the additive terms to sigmal
        sigmal += (3958.0 * sin(A1r) + 1962.0 * sin(Lprimer - Fr)
                   + 318.0 * sin(A2r))
        # Now use the tabla for sigmab
        sigmab = 0.0
        for i, value in enumerate(PERIODIC_TERMS_B_TABLE):
            argument = 0.0
            for j in range(4):
                if PERIODIC_TERMS_B_TABLE[i][j]:  # Avoid multiply by zero
                    argument += PERIODIC_TERMS_B_TABLE[i][j] * arguments[j]
            coeffb = value[4]
            if abs(value[1]) == 1:
                coeffb = coeffb * E
            elif abs(value[1]) == 2:
                coeffb = coeffb * E2
            sigmab += coeffb * sin(argument)
        # Add the additive terms to sigmab
        sigmab += (-2235.0 * sin(Lprimer) + 382.0 * sin(A3r)
                   + 175.0 * sin(A1r - Fr) + 175.0 * sin(A1r + Fr)
                   + 127.0 * sin(Lprimer - Mprimer)
                   - 115.0 * sin(Lprimer + Mprimer))
        Lambda = Lprime + (sigmal / 1000000.0)
        Beta = Angle(sigmab / 1000000.0)
        Delta = 385000.56 + (sigmar / 1000.0)
        ppii = asin(6378.14 / Delta)
        ppi = Angle(ppii, radians=True)
        return Lambda, Beta, Delta, ppi

    @staticmethod
    def apparent_ecliptical_pos(epoch):
        """This method computes the apparent geocentric ecliptical position
        (longitude, latitude) of the Moon for a given instant, referred to the
        mean equinox of the date, as well as the Moon-Earth distance in
        kilometers and the equatorial horizontal parallax.

        :param epoch: Instant to compute the Moon's position, as an
            py:class:`Epoch` object.
        :type epoch: :py:class:`Epoch`

        :returns: Tuple containing:

            * Apparent geocentric longitude of the center of the Moon, as an
              py:class:`Epoch` object.
            * Apparent geocentric latitude of the center of the Moon, as an
              py:class:`Epoch` object.
            * Distance in kilometers between the centers of Earth and Moon, in
              kilometers (float)
            * Equatorial horizontal parallax of the Moon, as an
              py:class:`Epoch` object.
        :rtype: tuple
        :raises: TypeError if input value is of wrong type.

        >>> epoch = Epoch(1992, 4, 12.0)
        >>> Lambda, Beta, Delta, ppi = Moon.apparent_ecliptical_pos(epoch)
        >>> print(round(Lambda, 5))
        133.16726
        >>> print(round(Beta, 6))
        -3.229126
        >>> print(round(Delta, 1))
        368409.7
        >>> print(round(ppi, 5))
        0.99199
        """

        # First check that input values are of correct types
        if not (isinstance(epoch, Epoch)):
            raise TypeError("Invalid input type")
        # Now, let's call the method geocentric_ecliptical_pos()
        Lambda, Beta, Delta, ppi = Moon.geocentric_ecliptical_pos(epoch)
        # Compute the nutation in longitude (deltaPsi)
        deltaPsi = nutation_longitude(epoch)
        # Correct the longitude to obtain the apparent longitude
        aLambda = Lambda + deltaPsi
        return aLambda, Beta, Delta, ppi

    @staticmethod
    def apparent_equatorial_pos(epoch):
        """This method computes the apparent equatorial position (right
        ascension, declination) of the Moon for a given instant, referred to
        the mean equinox of the date, as well as the Moon-Earth distance in
        kilometers and the equatorial horizontal parallax.

        :param epoch: Instant to compute the Moon's position, as an
            py:class:`Epoch` object.
        :type epoch: :py:class:`Epoch`

        :returns: Tuple containing:

            * Apparent right ascension of the center of the Moon, as an
              py:class:`Epoch` object.
            * Apparent declination of the center of the Moon, as an
              py:class:`Epoch` object.
            * Distance in kilometers between the centers of Earth and Moon, in
              kilometers (float)
            * Equatorial horizontal parallax of the Moon, as an
              py:class:`Epoch` object.
        :rtype: tuple
        :raises: TypeError if input value is of wrong type.

        >>> epoch = Epoch(1992, 4, 12.0)
        >>> ra, dec, Delta, ppi = Moon.apparent_equatorial_pos(epoch)
        >>> print(round(ra, 6))
        134.688469
        >>> print(round(dec, 6))
        13.768367
        >>> print(round(Delta, 1))
        368409.7
        >>> print(round(ppi, 5))
        0.99199
        """

        # First check that input values are of correct types
        if not (isinstance(epoch, Epoch)):
            raise TypeError("Invalid input type")
        # Let's start calling the method 'apparent_ecliptical_pos()'
        Lambda, Beta, Delta, ppi = Moon.apparent_ecliptical_pos(epoch)
        # Now we need the obliquity of the ecliptic
        epsilon = true_obliquity(epoch)
        # And now let's carry out the transformation ecliptical->equatorial
        ra, dec = ecliptical2equatorial(Lambda, Beta, epsilon)
        return ra, dec, Delta, ppi

    @staticmethod
    def longitude_mean_ascending_node(epoch):
        """This method computes the longitude of the mean ascending node of the
        Moon in degrees, for a given instant, measured from the mean equinox of
        the date.

        :param epoch: Instant to compute the Moon's mean ascending node, as an
            py:class:`Epoch` object.
        :type epoch: :py:class:`Epoch`

        :returns: The longitude of the mean ascending node.
        :rtype: py:class:`Angle`
        :raises: TypeError if input value is of wrong type.

        >>> epoch = Epoch(1913, 5, 27.0)
        >>> Omega = Moon.longitude_mean_ascending_node(epoch)
        >>> print(round(Omega, 1))
        0.0
        >>> epoch = Epoch(2043, 9, 10.0)
        >>> Omega = Moon.longitude_mean_ascending_node(epoch)
        >>> print(round(Omega, 1))
        0.0
        >>> epoch = Epoch(1959, 12, 7.0)
        >>> Omega = Moon.longitude_mean_ascending_node(epoch)
        >>> print(round(Omega, 1))
        180.0
        >>> epoch = Epoch(2108, 11, 3.0)
        >>> Omega = Moon.longitude_mean_ascending_node(epoch)
        >>> print(round(Omega, 1))
        180.0
        """

        # First check that input values are of correct types
        if not (isinstance(epoch, Epoch)):
            raise TypeError("Invalid input type")
        # Get the time from J2000.0 in Julian centuries
        t = (epoch - JDE2000) / 36525.0
        # Compute Moon's longitude of the mean ascending node
        Omega = 125.0445479 + (-1934.1362891
                               + (0.0020754
                                  + (1.0/476441.0
                                     - t/60616000.0) * t) * t) * t
        Omega = Angle(Omega).to_positive()
        return Omega

    @staticmethod
    def longitude_true_ascending_node(epoch):
        """This method computes the longitude of the true ascending node of the
        Moon in degrees, for a given instant, measured from the mean equinox of
        the date.

        :param epoch: Instant to compute the Moon's true ascending node, as an
            py:class:`Epoch` object.
        :type epoch: :py:class:`Epoch`

        :returns: The longitude of the true ascending node.
        :rtype: py:class:`Angle`
        :raises: TypeError if input value is of wrong type.

        >>> epoch = Epoch(1913, 5, 27.0)
        >>> Omega = Moon.longitude_true_ascending_node(epoch)
        >>> print(round(Omega, 4))
        0.8763
        """

        # First check that input values are of correct types
        if not (isinstance(epoch, Epoch)):
            raise TypeError("Invalid input type")
        # Let's start computing the longitude of the MEAN ascending node
        Omega = Moon.longitude_mean_ascending_node(epoch)
        # Get the time from J2000.0 in Julian centuries
        t = (epoch - JDE2000) / 36525.0
        # Mean elongation of the Moon
        D = 297.8501921 + (445267.1114034
                           + (-0.0018819
                              + (1.0/545868.0 - t/113065000.0) * t) * t) * t
        # Sun's mean anomaly
        M = 357.5291092 + (35999.0502909 + (-0.0001536 + t/24490000.0) * t) * t
        # Moon's mean anomaly
        Mprime = 134.9633964 + (477198.8675055
                                + (0.0087414
                                   + (1.0/69699.9
                                      + t/14712000.0) * t) * t) * t
        # Moon's argument of latitude
        F = 93.2720950 + (483202.0175233
                          + (-0.0036539
                             + (-1.0/3526000.0 + t/863310000.0) * t) * t) * t
        # Reduce the angles to a [0 360] range
        D = Angle(Angle.reduce_deg(D)).to_positive()
        Dr = D.rad()
        M = Angle(Angle.reduce_deg(M)).to_positive()
        Mr = M.rad()
        Mprime = Angle(Angle.reduce_deg(Mprime)).to_positive()
        Mprimer = Mprime.rad()
        F = Angle(Angle.reduce_deg(F)).to_positive()
        Fr = F.rad()
        # Compute the periodic terms
        corr = (-1.4979 * sin(2.0 * (Dr - Fr)) - 0.15 * sin(Mr)
                - 0.1226 * sin(2.0 * Dr) + 0.1176 * sin(2.0 * Fr)
                - 0.0801 * sin(2.0 * (Mprimer - Fr)))
        Omega += Angle(corr)
        return Omega

    @staticmethod
    def longitude_mean_perigee(epoch):
        """This method computes the longitude of the mean perigee of the lunar
        orbitn in degrees, for a given instant, measured from the mean equinox
        of the date.

        :param epoch: Instant to compute the Moon's mean perigee, as an
            py:class:`Epoch` object.
        :type epoch: :py:class:`Epoch`

        :returns: The longitude of the mean perigee.
        :rtype: py:class:`Angle`
        :raises: TypeError if input value is of wrong type.

        >>> epoch = Epoch(2021, 3, 5.0)
        >>> Pi = Moon.longitude_mean_perigee(epoch)
        >>> print(round(Pi, 5))
        224.89194
        """

        # First check that input values are of correct types
        if not (isinstance(epoch, Epoch)):
            raise TypeError("Invalid input type")
        # Get the time from J2000.0 in Julian centuries
        t = (epoch - JDE2000) / 36525.0
        # Compute Moon's longitude of the mean perigee
        ppii = 83.3532465 + (4069.0137287
                             + (-0.01032
                                + (-1.0/80053.0 + t/18999000.0) * t) * t) * t
        ppii = Angle(ppii)
        return ppii

    @staticmethod
    def illuminated_fraction_disk(epoch):
        """This method computes the approximate illuminated fraction 'k' of the
        disk of the Moon. The method used has a relatively low accuracy, but it
        is enough to the 2nd decimal place.

        :param epoch: Instant to compute the Moon's illuminated fraction of the
            disk, as a py:class:`Epoch` object.
        :type epoch: :py:class:`Epoch`

        :returns: The approximate illuminated fraction of the Moon's disk.
        :rtype: float
        :raises: TypeError if input value is of wrong type.

        >>> epoch = Epoch(1992, 4, 12.0)
        >>> k = Moon.illuminated_fraction_disk(epoch)
        >>> print(round(k, 2))
        0.68
        """

        # First check that input values are of correct types
        if not (isinstance(epoch, Epoch)):
            raise TypeError("Invalid input type")
        # Get the time from J2000.0 in Julian centuries
        t = (epoch - JDE2000) / 36525.0
        # Mean elongation of the Moon
        D = 297.8501921 + (445267.1114034
                           + (-0.0018819
                              + (1.0/545868.0 - t/113065000.0) * t) * t) * t
        # Sun's mean anomaly
        M = 357.5291092 + (35999.0502909 + (-0.0001536 + t/24490000.0) * t) * t
        # Moon's mean anomaly
        Mprime = 134.9633964 + (477198.8675055
                                + (0.0087414
                                   + (1.0/69699.9
                                      + t/14712000.0) * t) * t) * t
        # Reduce the angles to a [0 360] range
        D = Angle(Angle.reduce_deg(D)).to_positive()
        Dr = D.rad()
        M = Angle(Angle.reduce_deg(M)).to_positive()
        Mr = M.rad()
        Mprime = Angle(Angle.reduce_deg(Mprime)).to_positive()
        Mprimer = Mprime.rad()
        # Compute the 'i' angle
        i = Angle(180.0 - D - 6.289 * sin(Mprimer) + 2.1 * sin(Mr)
                  - 1.274 * sin(2.0 * Dr - Mprimer) - 0.658 * sin(2.0 * Dr)
                  - 0.214 * sin(2.0 * Mprimer) - 0.11 * sin(Dr))
        k = (1.0 + cos(i.rad())) / 2.0
        return k

    @staticmethod
    def position_bright_limb(epoch):
        """This method computes the position angle of the Moon's bright limb,
        i.e., the position angle of the midpoint of the illuminated limb,
        reckoned eastward from the North Point of the disk (not from the axis
        of rotation of the lunar globe).

        :param epoch: Instant to compute the position angle of the  Moon's
            bright limb, as a py:class:`Epoch` object.
        :type epoch: :py:class:`Epoch`

        :returns: The position angle of the Moon's bright limb.
        :rtype: :py:class:`Angle`
        :raises: TypeError if input value is of wrong type.

        >>> epoch = Epoch(1992, 4, 12.0)
        >>> xi = Moon.position_bright_limb(epoch)
        >>> print(round(xi, 1))
        285.0
        """

        # First check that input values are of correct types
        if not (isinstance(epoch, Epoch)):
            raise TypeError("Invalid input type")
        # Compute the right ascension and declination of the Sun
        a0, d0, r0 = Sun.apparent_rightascension_declination_coarse(epoch)
        # Now compute the right ascension and declination of the Moon
        a, d, r, ppi = Moon.apparent_equatorial_pos(epoch)
        a0r = a0.rad()
        d0r = d0.rad()
        ar = a.rad()
        dr = d.rad()
        # Compute the numerator of the tan(xi) formula
        numerator = cos(d0r) * sin(a0r - ar)
        # Now the denominator
        denominator = sin(d0r) * cos(dr) - cos(d0r) * sin(dr) * cos(a0r - ar)
        # Now let's compute xi
        xi = atan2(numerator, denominator)
        xi = Angle(xi, radians=True).to_positive()
        return xi

    @staticmethod
    def moon_phase(epoch, target="new"):
        """This method computes the time of the phase of the moon closest to
        the provided epoch. The resulting time is expressed in the uniform time
        scale of Dynamical Time (TT).

        :param epoch: Approximate epoch we want to compute the Moon phase for.
        :type year: :py:class:`Epoch`
        :param target: Corresponding phase. It can be "new" (New Moon), "first"
            (First Quarter), "full" (Full Moon) and "last" (Last Quarter). It
            is 'new' by default.
        :type target: str

        :returns: The instant of time when the provided phase happens.
        :rtype: :py:class:`Epoch`
        :raises: TypeError if input values are of wrong type.
        :raises: ValueError if 'target' value is invalid.

        >>> epoch = Epoch(1977, 2, 15.0)
        >>> new_moon = Moon.moon_phase(epoch, target="new")
        >>> y, m, d, h, mi, s = new_moon.get_full_date()
        >>> print("{}/{}/{} {}:{}:{}".format(y, m, d, h, mi, round(s, 0)))
        1977/2/18 3:37:42.0
        >>> epoch = Epoch(2044, 1, 1.0)
        >>> new_moon = Moon.moon_phase(epoch, target="last")
        >>> y, m, d, h, mi, s = new_moon.get_full_date()
        >>> print("{}/{}/{} {}:{}:{}".format(y, m, d, h, mi, round(s)))
        2044/1/21 23:48:17
        """

        # First check that input values are of correct types
        if not (isinstance(epoch, Epoch) and isinstance(target, str)):
            raise TypeError("Invalid input types")
        # Second, check that the target is correct
        if (
            (target != "new")
            and (target != "first")
            and (target != "full")
            and (target != "last")
        ):
            raise ValueError("'target' value is invalid")
        # Let's start computing the year with decimals
        y, m, d = epoch.get_date()
        num_days_year = 365.0
        if Epoch.is_leap(y):
            num_days_year = 366.0
        doy = Epoch.get_doy(y, m, d)
        year = y + doy / num_days_year
        # We compute the 'k' parameter
        k = round((year - 2000.0) * 12.3685, 0)
        if target == "first":
            k += 0.25
        elif target == "full":
            k += 0.5
        elif target == "last":
            k += 0.75
        t = k / 1236.85
        # Compute the time of the 'mean' phase of the Moon
        jde = (2451550.09766 + 29.530588861 * k
               + (0.00015437 + (-0.00000015 + 0.00000000073 * t) * t) * t * t)
        # Eccentricity of Earth's orbit around the Sun
        E = 1.0 + (-0.002516 - 0.0000074 * t) * t
        # Sun's mean anomaly
        M = 2.5534 + 29.1053567 * k + (-0.0000014 - 0.00000011 * t) * t * t
        # Moon's mean anomaly
        Mprime = (201.5643 + 385.81693528 * k
                  + (0.0107582 + (0.00001238 - 0.000000058 * t) * t) * t * t)
        # Moon's argument of latitude
        F = (160.7108 + 390.67050284 * k
             + (-0.0016118 + (-0.00000227 + 0.000000011 * t) * t) * t * t)
        # Longitude of the ascending node of the lunar orbit
        Omega = (124.7746 - 1.56375588 * k
                 + (0.0020672 + 0.00000215 * t) * t * t)
        M = Angle(Angle.reduce_deg(M)).to_positive()
        Mr = M.rad()
        Mprime = Angle(Angle.reduce_deg(Mprime)).to_positive()
        Mprimer = Mprime.rad()
        F = Angle(Angle.reduce_deg(F)).to_positive()
        Fr = F.rad()
        Omega = Angle(Angle.reduce_deg(Omega)).to_positive()
        Omegar = Omega.rad()
        # Planetary arguments
        a1 = 299.77 + 0.107408 * k - 0.009173 * t * t
        a2 = 251.88 + 0.016321 * k
        a3 = 251.83 + 26.651886 * k
        a4 = 349.42 + 36.412478 * k
        a5 = 84.66 + 18.206239 * k
        a6 = 141.74 + 53.303771 * k
        a7 = 207.14 + 2.453732 * k
        a8 = 154.84 + 7.30686 * k
        a9 = 34.52 + 27.261239 * k
        a10 = 207.19 + 0.121824 * k
        a11 = 291.34 + 1.844379 * k
        a12 = 161.72 + 24.198154 * k
        a13 = 239.56 + 25.513099 * k
        a14 = 331.55 + 3.592518 * k
        a1 = Angle(Angle.reduce_deg(a1)).to_positive()
        a1r = a1.rad()
        a2 = Angle(Angle.reduce_deg(a2)).to_positive()
        a2r = a2.rad()
        a3 = Angle(Angle.reduce_deg(a3)).to_positive()
        a3r = a3.rad()
        a4 = Angle(Angle.reduce_deg(a4)).to_positive()
        a4r = a4.rad()
        a5 = Angle(Angle.reduce_deg(a5)).to_positive()
        a5r = a5.rad()
        a6 = Angle(Angle.reduce_deg(a6)).to_positive()
        a6r = a6.rad()
        a7 = Angle(Angle.reduce_deg(a7)).to_positive()
        a7r = a7.rad()
        a8 = Angle(Angle.reduce_deg(a8)).to_positive()
        a8r = a8.rad()
        a9 = Angle(Angle.reduce_deg(a9)).to_positive()
        a9r = a9.rad()
        a10 = Angle(Angle.reduce_deg(a10)).to_positive()
        a10r = a10.rad()
        a11 = Angle(Angle.reduce_deg(a11)).to_positive()
        a11r = a11.rad()
        a12 = Angle(Angle.reduce_deg(a12)).to_positive()
        a12r = a12.rad()
        a13 = Angle(Angle.reduce_deg(a13)).to_positive()
        a13r = a13.rad()
        a14 = Angle(Angle.reduce_deg(a14)).to_positive()
        a14r = a14.rad()
        # Now let's compute the corrections
        corr = 0.0
        w = 0.0
        if target == "new":
            corr = (-0.4072 * sin(Mprimer) + 0.17241 * E * sin(Mr)
                    + 0.01608 * sin(2.0 * Mprimer) + 0.01039 * sin(2.0 * Fr)
                    + 0.00739 * E * sin(Mprimer - Mr)
                    - 0.00514 * E * sin(Mprimer + Mr)
                    + 0.00208 * E * E * sin(2.0 * Mr)
                    - 0.00111 * sin(Mprimer - 2.0 * Fr)
                    - 0.00057 * sin(Mprimer + 2.0 * Fr)
                    + 0.00056 * E * sin(2.0 * Mprimer + Mr)
                    - 0.00042 * sin(3.0 * Mprimer)
                    + 0.00042 * E * sin(Mr + 2.0 * Fr)
                    + 0.00038 * E * sin(Mr - 2.0 * Fr)
                    - 0.00024 * E * sin(2.0 * Mprimer - Mr)
                    - 0.00017 * sin(Omegar) - 0.00007 * sin(Mprimer + 2.0 * Mr)
                    + 0.00004 * sin(2.0 * (Mprimer - Fr))
                    + 0.00004 * sin(3.0 * Mr)
                    + 0.00003 * sin(Mprimer + Mr - 2.0 * Fr)
                    + 0.00003 * sin(2.0 * (Mprimer + Fr))
                    - 0.00003 * sin(Mprimer + Mr + 2.0 * Fr)
                    + 0.00003 * sin(Mprimer - Mr + 2.0 * Fr)
                    - 0.00002 * sin(Mprimer - Mr - 2.0 * Fr)
                    - 0.00002 * sin(3.0 * Mprimer + Mr)
                    + 0.00002 * sin(4.0 * Mprimer))
        elif target == "full":
            corr = (-0.40614 * sin(Mprimer) + 0.17302 * E * sin(Mr)
                    + 0.01614 * sin(2.0 * Mprimer) + 0.01043 * sin(2.0 * Fr)
                    + 0.00734 * E * sin(Mprimer - Mr)
                    - 0.00515 * E * sin(Mprimer + Mr)
                    + 0.00209 * E * E * sin(2.0 * Mr)
                    - 0.00111 * sin(Mprimer - 2.0 * Fr)
                    - 0.00057 * sin(Mprimer + 2.0 * Fr)
                    + 0.00056 * E * sin(2.0 * Mprimer + Mr)
                    - 0.00042 * sin(3.0 * Mprimer)
                    + 0.00042 * E * sin(Mr + 2.0 * Fr)
                    + 0.00038 * E * sin(Mr - 2.0 * Fr)
                    - 0.00024 * E * sin(2.0 * Mprimer - Mr)
                    - 0.00017 * sin(Omegar) - 0.00007 * sin(Mprimer + 2.0 * Mr)
                    + 0.00004 * sin(2.0 * (Mprimer - Fr))
                    + 0.00004 * sin(3.0 * Mr)
                    + 0.00003 * sin(Mprimer + Mr - 2.0 * Fr)
                    + 0.00003 * sin(2.0 * (Mprimer + Fr))
                    - 0.00003 * sin(Mprimer + Mr + 2.0 * Fr)
                    + 0.00003 * sin(Mprimer - Mr + 2.0 * Fr)
                    - 0.00002 * sin(Mprimer - Mr - 2.0 * Fr)
                    - 0.00002 * sin(3.0 * Mprimer + Mr)
                    + 0.00002 * sin(4.0 * Mprimer))
        elif target == "first" or target == "last":
            corr = (-0.62801 * sin(Mprimer) + 0.17172 * E * sin(Mr)
                    - 0.01183 * E * sin(Mprimer + Mr)
                    + 0.00862 * sin(2.0 * Mprimer) + 0.00804 * sin(2.0 * Fr)
                    + 0.00454 * E * sin(Mprimer - Mr)
                    + 0.00204 * E * E * sin(2.0 * Mr)
                    - 0.0018 * sin(Mprimer - 2.0 * Fr)
                    - 0.0007 * sin(Mprimer + 2.0 * Fr)
                    - 0.0004 * sin(3.0 * Mprimer)
                    - 0.00034 * E * sin(2.0 * Mprimer - Mr)
                    + 0.00032 * E * sin(Mr + 2.0 * Fr)
                    + 0.00032 * E * sin(Mr - 2.0 * Fr)
                    - 0.00028 * E * E * sin(Mprimer + 2.0 * Mr)
                    + 0.00027 * E * sin(2.0 * Mprimer + Mr)
                    - 0.00017 * sin(Omegar)
                    - 0.00005 * sin(Mprimer - Mr - 2.0 * Fr)
                    + 0.00004 * sin(2.0 * (Mprimer + Fr))
                    - 0.00004 * sin(Mprimer + Mr + 2.0 * Fr)
                    + 0.00004 * sin(Mprimer - 2.0 * Mr)
                    + 0.00003 * sin(Mprimer + Mr - 2.0 * Fr)
                    + 0.00003 * sin(3.0 * Mr)
                    + 0.00002 * sin(2.0 * (Mprimer - Fr))
                    + 0.00002 * sin(Mprimer - Mr + 2.0 * Fr)
                    - 0.00002 * sin(3.0 * Mprimer + Mr))
            w = (0.00306 - 0.00038 * E * cos(Mr) + 0.00026 * cos(Mprimer)
                 - 0.00002 * cos(Mprimer - Mr) + 0.00002 * cos(Mprimer + Mr)
                 + 0.00002 * cos(2.0 * Fr))
            if target == "last":
                w = -w
        # Additional corrections for all phases
        corr2 = (0.000325 * sin(a1r) + 0.000165 * sin(a2r)
                 + 0.000164 * sin(a3r) + 0.000126 * sin(a4r)
                 + 0.000110 * sin(a5r) + 0.000062 * sin(a6r)
                 + 0.000060 * sin(a7r) + 0.000056 * sin(a8r)
                 + 0.000047 * sin(a9r) + 0.000042 * sin(a10r)
                 + 0.000040 * sin(a11r) + 0.000037 * sin(a12r)
                 + 0.000035 * sin(a13r) + 0.000023 * sin(a14r))
        jde += corr + corr2 + w
        jde = Epoch(jde)
        return jde

    @staticmethod
    def moon_perigee_apogee(epoch, target="perigee"):
        """This method computes the approximate times when the distance between
        the Earth and the Moon is a minimum (perigee) or a maximum (apogee).
        The resulting times will be expressed in the uniform time scale of
        Dynamical Time (TT).

        :param epoch: Approximate epoch we want to compute the Moon's perigee
            or apogee for.
        :type year: :py:class:`Epoch`
        :param target: Either 'perigee' or 'apogee'. It's 'perigee' by default.
        :type target: str

        :returns: A tuple containing the instant of time when the perigee or
            apogee happens, as a :py:class:`Epoch` object, and the Moon's
            corresponding equatorial horizontal parallax, as a
            :py:class:`Angle` object.
        :rtype: tuple
        :raises: TypeError if input values are of wrong type.
        :raises: ValueError if 'target' value is invalid.

        >>> epoch = Epoch(1988, 10, 1.0)
        >>> apogee, parallax = Moon.moon_perigee_apogee(epoch, target="apogee")
        >>> y, m, d, h, mi, s = apogee.get_full_date()
        >>> print("{}/{}/{} {}:{}".format(y, m, d, h, mi))
        1988/10/7 20:30
        >>> print("{}".format(parallax.dms_str(n_dec=3)))
        54' 0.679''
        """

        # First check that input values are of correct types
        if not (isinstance(epoch, Epoch) and isinstance(target, str)):
            raise TypeError("Invalid input types")
        # Second, check that the target is correct
        if (
            (target != "perigee")
            and (target != "apogee")
        ):
            raise ValueError("'target' value is invalid")
        # Let's start computing the year with decimals
        y, m, d = epoch.get_date()
        num_days_year = 365.0
        if Epoch.is_leap(y):
            num_days_year = 366.0
        doy = Epoch.get_doy(y, m, d)
        year = y + doy / num_days_year
        # We compute the 'k' parameter
        k = round((year - 1999.97) * 13.2555, 0)
        if target == "apogee":
            k += 0.5
        t = k / 1325.55
        # Compute the time of the 'mean' perigee or apogee
        jde = (2451534.6698 + 27.55454989 * k
               + (-0.0006691 + (0.000001098 + 0.0000000052 * t) * t) * t * t)
        # Moon's mean elongation at jde
        D = (171.9179 + 335.9106046 * k
             + (-0.0100383 + (-0.00001156 + 0.000000055 * t) * t) * t * t)
        # Sun's mean anomaly
        M = 347.3477 + 27.1577721 * k + (-0.000813 - 0.000001 * t) * t * t
        # Moon's argument of latitude
        F = 316.6109 + 364.5287911 * k + (-0.0125053 - 0.0000148 * t) * t * t
        D = Angle(Angle.reduce_deg(D)).to_positive()
        Dr = D.rad()
        M = Angle(Angle.reduce_deg(M)).to_positive()
        Mr = M.rad()
        F = Angle(Angle.reduce_deg(F)).to_positive()
        Fr = F.rad()
        corr = 0.0
        parallax = 0.0
        if target == "perigee":
            corr = (-1.6769 * sin(2.0 * Dr) + 0.4589 * sin(4.0 * Dr)
                    - 0.1856 * sin(6.0 * Dr) + 0.0883 * sin(8.0 * Dr)
                    + (-0.0773 + 0.00019 * t) * sin(2.0 * Dr - Mr)
                    + (0.0502 - 0.00013 * t) * sin(Mr) - 0.046 * sin(10.0 * Dr)
                    + (0.0422 - 0.00011 * t) * sin(4.0 * Dr - Mr)
                    - 0.0256 * sin(6.0 * Dr - Mr) + 0.0253 * sin(12.0 * Dr)
                    + 0.0237 * sin(Dr) + 0.0162 * sin(8.0 * Dr - Mr)
                    - 0.0145 * sin(14.0 * Dr) + 0.0129 * sin(2.0 * Fr)
                    - 0.0112 * sin(3.0 * Dr) - 0.0104 * sin(10.0 * Dr - Mr)
                    + 0.0086 * sin(16.0 * Dr) + 0.0069 * sin(12.0 * Dr - Mr)
                    + 0.0066 * sin(5.0 * Dr) - 0.0053 * sin(2.0 * (Dr + Fr))
                    - 0.0052 * sin(18.0 * Dr) - 0.0046 * sin(14.0 * Dr - Mr)
                    - 0.0041 * sin(7.0 * Dr) + 0.004 * sin(2.0 * Dr + Mr)
                    + 0.0032 * sin(20.0 * Dr) - 0.0032 * sin(Dr + Mr)
                    + 0.0031 * sin(16.0 * Dr - Mr)
                    - 0.0029 * sin(4.0 * Dr + Mr) + 0.0027 * sin(9.0 * Dr)
                    + 0.0027 * sin(4.0 * Dr + 2.0 * Fr)
                    - 0.0027 * sin(2.0 * (Dr - Mr))
                    + 0.0024 * sin(4.0 * Dr - 2.0 * Mr)
                    - 0.0021 * sin(6.0 * Dr - 2.0 * Mr)
                    - 0.0021 * sin(22.0 * Dr) - 0.0021 * sin(18.0 * Dr - Mr)
                    + 0.0019 * sin(6.0 * Dr + Mr) - 0.0018 * sin(11.0 * Dr)
                    - 0.0014 * sin(8.0 * Dr + Mr)
                    - 0.0014 * sin(4.0 * Dr - 2.0 * Fr)
                    - 0.0014 * sin(6.0 * Dr + 2.0 * Fr)
                    + 0.0014 * sin(3.0 * Dr + Mr) - 0.0014 * sin(5.0 * Dr + Mr)
                    + 0.0013 * sin(13.0 * Dr) + 0.0013 * sin(20.0 * Dr - Mr)
                    + 0.0011 * sin(3.0 * Dr + 2.0 * Mr)
                    - 0.0011 * sin(4.0 * Dr + 2.0 * Fr - 2.0 * Mr)
                    - 0.0010 * sin(Dr + 2.0 * Mr)
                    - 0.0009 * sin(22.0 * Dr - Mr) - 0.0008 * sin(4.0 * Fr)
                    + 0.0008 * sin(6.0 * Dr - 2.0 * Fr)
                    + 0.0008 * sin(2.0 * Dr - 2.0 * Fr + Mr)
                    + 0.0007 * sin(2.0 * Mr) + 0.0007 * sin(2.0 * Fr - Mr)
                    + 0.0007 * sin(2.0 * Dr + 4.0 * Fr)
                    - 0.0006 * sin(2.0 * (Fr - Mr))
                    - 0.0006 * sin(2.0 * (Dr - Fr + Mr))
                    + 0.0006 * sin(24.0 * Dr) + 0.0005 * sin(4.0 * (Dr - Fr))
                    + 0.0005 * sin(2.0 * (Dr + Mr)) - 0.0004 * sin(Dr - Mr))
            parallax = (3629.215 + 63.224 * cos(2.0 * Dr)
                        - 6.99 * cos(4.0 * Dr)
                        + (2.834 - 0.0071 * t) * cos(2.0 * Dr - Mr)
                        + 1.927 * cos(6.0 * Dr) - 1.263 * cos(Dr)
                        - 0.702 * cos(8.0 * Dr)
                        + (0.696 - 0.0017 * t) * cos(Mr) - 0.69 * cos(2.0 * Fr)
                        + (-0.629 + 0.0016 * t) * cos(4.0 * Dr - Mr)
                        - 0.392 * cos(2.0 * (Dr - Fr)) + 0.297 * cos(10.0 * Dr)
                        + 0.26 * cos(6.0 * Dr - Mr) + 0.201 * cos(3.0 * Dr)
                        - 0.161 * cos(2.0 * Dr + Mr) + 0.157 * cos(Dr + Mr)
                        - 0.138 * cos(12.0 * Dr) - 0.127 * cos(8.0 * Dr - Mr)
                        + 0.104 * cos(2.0 * (Dr + Fr))
                        + 0.104 * cos(2.0 * (Dr - Mr)) - 0.079 * cos(5.0 * Dr)
                        + 0.068 * cos(14.0 * Dr) + 0.067 * cos(10.0 * Dr - Mr)
                        + 0.054 * cos(4.0 * Dr + Mr)
                        - 0.038 * cos(12.0 * Dr - Mr)
                        - 0.038 * cos(4.0 * Dr - 2.0 * Mr)
                        + 0.037 * cos(7.0 * Dr)
                        - 0.037 * cos(4.0 * Dr + 2.0 * Fr)
                        - 0.035 * cos(16.0 * Dr) - 0.03 * cos(3.0 * Dr + Mr)
                        + 0.029 * cos(Dr - Mr) - 0.025 * cos(6.0 * Dr + Mr)
                        + 0.023 * cos(2.0 * Mr) + 0.023 * cos(14.0 * Dr - Mr)
                        - 0.023 * cos(2.0 * (Dr + Mr))
                        + 0.022 * cos(6.0 * Dr - 2.0 * Mr)
                        - 0.021 * cos(2.0 * (Dr - Fr) - Mr)
                        - 0.020 * cos(9.0 * Dr) + 0.019 * cos(18.0 * Dr)
                        + 0.017 * cos(6.0 * Dr + 2.0 * Fr)
                        + 0.014 * cos(2.0 * Fr - Mr)
                        - 0.014 * cos(16.0 * Dr - Mr)
                        + 0.013 * cos(4.0 * Dr - 2.0 * Fr)
                        + 0.012 * cos(8.0 * Dr + Mr) + 0.011 * cos(11.0 * Dr)
                        + 0.01 * cos(5.0 * Dr + Mr) - 0.01 * cos(20.0 * Dr))
        else:
            corr = (0.4392 * sin(2.0 * Dr)
                    + 0.0684 * sin(4.0 * Dr)
                    + (0.0456 - 0.00011 * t) * sin(Mr)
                    + (0.0426 - 0.00011 * t) * sin(2.0 * Dr - Mr)
                    + 0.0212 * sin(2.0 * Fr)
                    - 0.0189 * sin(Dr)
                    + 0.0144 * sin(6.0 * Dr)
                    + 0.0113 * sin(4.0 * Dr - Mr)
                    + 0.0047 * sin(2.0 * (Dr + Fr))
                    + 0.0036 * sin(Dr + Mr)
                    + 0.0035 * sin(8.0 * Dr)
                    + 0.0034 * sin(6.0 * Dr - Mr)
                    - 0.0034 * sin(2.0 * (Dr - Fr))
                    + 0.0022 * sin(2.0 * (Dr - Mr))
                    - 0.0017 * sin(3.0 * Dr)
                    + 0.0013 * sin(4.0 * Dr + 2.0 * Fr)
                    + 0.0011 * sin(8.0 * Dr - Mr)
                    + 0.0010 * sin(4.0 * Dr - 2.0 * Mr)
                    + 0.0009 * sin(10.0 * Dr)
                    + 0.0007 * sin(3.0 * Dr + Mr)
                    + 0.0006 * sin(2.0 * Mr)
                    + 0.0005 * sin(2.0 * Dr + Mr)
                    + 0.0005 * sin(2.0 * (Dr + Mr))
                    + 0.0004 * sin(6.0 * Dr + 2.0 * Fr)
                    + 0.0004 * sin(6.0 * Dr - 2.0 * Mr)
                    + 0.0004 * sin(10.0 * Dr - Mr)
                    - 0.0004 * sin(5.0 * Dr)
                    - 0.0004 * sin(4.0 * Dr - 2.0 * Fr)
                    + 0.0003 * sin(2.0 * Fr + Mr)
                    + 0.0003 * sin(12.0 * Dr)
                    + 0.0003 * sin(2.0 * (Dr + Fr) - Mr)
                    - 0.0003 * sin(Dr - Mr))
            parallax = (3245.251 - 9.147 * cos(2.0 * Dr) - 0.841 * cos(Dr)
                        + 0.697 * cos(2.0 * Fr)
                        + (-0.656 + 0.0016 * t) * cos(Mr)
                        + 0.355 * cos(4.0 * Dr) + 0.159 * cos(2.0 * Dr - Mr)
                        + 0.127 * cos(Dr + Mr) + 0.065 * cos(4.0 * Dr - Mr)
                        + 0.052 * cos(6.0 * Dr) + 0.043 * cos(2.0 * Dr + Mr)
                        + 0.031 * cos(2.0 * (Dr + Fr))
                        - 0.023 * cos(2.0 * (Dr - Fr))
                        + 0.022 * cos(2.0 * (Dr - Mr))
                        + 0.019 * cos(2.0 * (Dr + Mr)) - 0.016 * cos(2.0 * Mr)
                        + 0.014 * cos(6.0 * Dr - Mr) + 0.01 * cos(8.0 * Dr))
        jde += corr
        jde = Epoch(jde)
        parallax = Angle(0, 0, parallax)
        return jde, parallax

    @staticmethod
    def moon_passage_nodes(epoch, target="ascending"):
        """This method computes the approximate times when the center of the
        Moon passes through the ascending or descending node of its orbit. The
        resulting times will be expressed in the uniform time scale of
        Dynamical Time (TT).

        :param epoch: Approximate epoch we want to compute the Moon's passage
            through the ascending or descending node.
        :type year: :py:class:`Epoch`
        :param target: Either 'ascending' or 'descending'. It is 'ascending' by
            default.
        :type target: str

        :returns: The instant of time when the Moon passes thhrough the
            ascending or descending node.
        :rtype: :py:class:`Epoch`
        :raises: TypeError if input values are of wrong type.
        :raises: ValueError if 'target' value is invalid.

        >>> epoch = Epoch(1987, 5, 15.0)
        >>> passage = Moon.moon_passage_nodes(epoch, target="ascending")
        >>> y, m, d, h, mi, s = passage.get_full_date()
        >>> mi += s/60.0
        >>> print("{}/{}/{} {}:{}".format(y, m, d, h, round(mi)))
        1987/5/23 6:26
        """

        # First check that input values are of correct types
        if not (isinstance(epoch, Epoch) and isinstance(target, str)):
            raise TypeError("Invalid input types")
        # Second, check that the target is correct
        if (
            (target != "ascending")
            and (target != "descending")
        ):
            raise ValueError("'target' value is invalid")
        # Let's start computing the year with decimals
        y, m, d = epoch.get_date()
        num_days_year = 365.0
        if Epoch.is_leap(y):
            num_days_year = 366.0
        doy = Epoch.get_doy(y, m, d)
        year = y + doy / num_days_year
        # Compute the 'k' parameter
        k = round((year - 2000.05) * 13.4223, 0)
        if target == "descending":
            k += 0.5
        t = k / 1342.23
        # Compute the time without the corrections
        jde = (2451565.1619 + 27.212220817 * k
               + (0.0002762 + (0.000000021 - 0.000000000088 * t) * t) * t * t)
        # Compute the following angles in degrees
        D = (183.638 + 331.73735682 * k
             + (0.0014852 + (0.00000209 - 0.00000001 * t) * t) * t * t)
        M = 17.4006 + 26.8203725 * k + (0.0001186 + 0.00000006 * t) * t * t
        Mprime = (38.3776 + 355.52747313 * k
                  + (0.0123499 + (0.000014627 - 0.000000069 * t) * t) * t * t)
        Omega = (123.9767 - 1.44098956 * k
                 + (0.0020608 + (0.00000214 - 0.000000016 * t) * t) * t * t)
        V = 299.75 + (132.85 - 0.009173 * t) * t
        P = Omega + 272.75 - 2.3 * t
        # Reduce the angles to the [0 360] range, and convert to radians
        D = Angle(Angle.reduce_deg(D)).to_positive()
        Dr = D.rad()
        M = Angle(Angle.reduce_deg(M)).to_positive()
        Mr = M.rad()
        Mprime = Angle(Angle.reduce_deg(Mprime)).to_positive()
        Mprimer = Mprime.rad()
        Omega = Angle(Angle.reduce_deg(Omega)).to_positive()
        Omegar = Omega.rad()
        V = Angle(Angle.reduce_deg(V)).to_positive()
        Vr = V.rad()
        P = Angle(Angle.reduce_deg(P)).to_positive()
        Pr = P.rad()
        # Eccentricity of Earth's orbit around the Sun
        E = 1.0 + (-0.002516 - 0.0000074 * t) * t
        # Compute the correction to jde
        corr = (-0.4721 * sin(Mprimer) - 0.1649 * sin(2.0 * Dr)
                - 0.0868 * sin(2.0 * Dr - Mprimer)
                + 0.0084 * sin(2.0 * Dr + Mprimer)
                - 0.0083 * E * sin(2.0 * Dr - Mr)
                - 0.0039 * E * sin(2.0 * Dr - Mr - Mprimer)
                + 0.0034 * sin(2.0 * Mprimer)
                - 0.0031 * sin(2.0 * (Dr - Mprimer))
                + 0.0030 * E * sin(2.0 * Dr + Mr)
                + 0.0028 * E * sin(Mr - Mprimer) + 0.0026 * E * sin(Mr)
                + 0.0025 * sin(4.0 * Dr) + 0.0024 * sin(Dr)
                + 0.0022 * E * sin(Mr + Mprimer) + 0.0017 * sin(Omegar)
                + 0.0014 * sin(4.0 * Dr - Mprimer)
                + 0.0005 * E * sin(2.0 * Dr + Mr - Mprimer)
                + 0.0004 * E * sin(2.0 * Dr - Mr + Mprimer)
                - 0.0003 * E * sin(2.0 * (Dr - Mr))
                + 0.0003 * E * sin(4.0 * Dr - Mr) + 0.0003 * sin(Vr)
                + 0.0003 * sin(Pr))
        jde += corr
        jde = Epoch(jde)
        return jde

    @staticmethod
    def moon_maximum_declination(epoch, target="northern"):
        """This method computes the approximate times when the Moon reaches
        its maximum declination (either 'northern' or 'southern'), as well as
        the values of these extreme declinations. The resulting times will be
        expressed in the uniform time scale of Dynamical Time (TT).

        :param epoch: Approximate epoch we want to compute the Moon's maximum
            declination.
        :type year: :py:class:`Epoch`
        :param target: Either 'northern' or 'southern', depending on the
            maximum declination being looked for. It is 'northern' by default.
        :type target: str

        :returns: A tuple containing the instant of time when the maximum
            declination happens, as a :py:class:`Epoch` object, and the angle
            value of such declination, as a :py:class:`Angle` object.
        :rtype: tuple
        :raises: TypeError if input value is of wrong type.

        >>> epoch = Epoch(1988, 12, 15.0)
        >>> epo, dec = Moon.moon_maximum_declination(epoch)
        >>> y, m, d, h, mi, s = epo.get_full_date()
        >>> print("{}/{}/{} {}:0{}".format(y, m, d, h, mi))
        1988/12/22 20:01
        >>> print("{}".format(dec.dms_str(n_dec=0)))
        28d 9' 22.0''
        >>> epoch = Epoch(2049, 4, 15.0)
        >>> epo, dec = Moon.moon_maximum_declination(epoch, target='southern')
        >>> y, m, d, h, mi, s = epo.get_full_date()
        >>> print("{}/{}/{} {}:{}".format(y, m, d, h, mi))
        2049/4/21 14:0
        >>> print("{}".format(dec.dms_str(n_dec=0)))
        -22d 8' 18.0''
        >>> epoch = Epoch(-4, 3, 15.0)
        >>> epo, dec = Moon.moon_maximum_declination(epoch, target='northern')
        >>> y, m, d, h, mi, s = epo.get_full_date()
        >>> print("{}/{}/{} {}h".format(y, m, d, h))
        -4/3/16 15h
        >>> print("{}".format(dec.dms_str(n_dec=0)))
        28d 58' 26.0''
        """

        # First check that input values are of correct types
        if not (isinstance(epoch, Epoch) and isinstance(target, str)):
            raise TypeError("Invalid input types")
        # Second, check that the target is correct
        if (
            (target != "northern")
            and (target != "southern")
        ):
            raise ValueError("'target' value is invalid")
        # Let's start computing the year with decimals
        y, m, d = epoch.get_date()
        num_days_year = 365.0
        if Epoch.is_leap(y):
            num_days_year = 366.0
        doy = Epoch.get_doy(y, m, d)
        year = y + doy / num_days_year
        # We compute the 'k' parameter
        k = round((year - 2000.03) * 13.3686, 0)
        t = k / 1336.86
        # Compute the following angles in degrees, plus 'jde' in days
        D = 333.0705546 * k + (-0.0004214 + 0.00000011 * t) * t * t
        M = 26.9281592 * k - (0.0000355 + 0.0000001 * t) * t * t
        Mprime = 356.9562794 * k + (0.0103066 + 0.00001251 * t) * t * t
        F = 1.4467807 * k - (0.002069 + 0.00000215 * t) * t * t
        jde = 27.321582247 * k + (0.000119804 - 0.000000141 * t) * t * t
        # Adjust the values according to northern of southern declination
        if (target == 'northern'):
            D += 152.2029
            M += 14.8591
            Mprime += 4.6881
            F += 325.8867
            jde += 2451562.5897
        else:
            D += 345.6676
            M += 1.13951
            Mprime += 186.21
            F += 145.1633
            jde += 2451548.9289
        # Reduce the angles to the [0 360] range, and convert to radians
        D = Angle(Angle.reduce_deg(D)).to_positive()
        Dr = D.rad()
        M = Angle(Angle.reduce_deg(M)).to_positive()
        Mr = M.rad()
        Mprime = Angle(Angle.reduce_deg(Mprime)).to_positive()
        Mprimer = Mprime.rad()
        F = Angle(Angle.reduce_deg(F)).to_positive()
        Fr = F.rad()
        # Eccentricity of Earth's orbit around the Sun
        E = 1.0 + (-0.002516 - 0.0000074 * t) * t
        corr = 0.0
        cor2 = 0.0
        # Compute the periodic terms for the time of maximum declination
        if (target == 'northern'):
            # Correction for the epoch
            corr = (0.8975 * cos(Fr) - 0.4726 * sin(Mprimer)
                    - 0.1030 * sin(2.0 * Fr) - 0.0976 * sin(2.0 * Dr - Mprimer)
                    - 0.0462 * cos(Mprimer - Fr) - 0.0461 * cos(Mprimer + Fr)
                    - 0.0438 * sin(2.0 * Dr) + 0.0162 * E * sin(Mr)
                    - 0.0157 * cos(3.0 * Fr) + 0.0145 * sin(Mprimer + 2.0 * Fr)
                    + 0.0136 * cos(2.0 * Dr - Fr)
                    - 0.0095 * cos(2.0 * Dr - Mprimer - Fr)
                    - 0.0091 * cos(2.0 * Dr - Mprimer + Fr)
                    - 0.0089 * cos(2.0 * Dr + Fr) + 0.0075 * sin(2.0 * Mprimer)
                    - 0.0068 * sin(Mprimer - 2.0 * Fr)
                    + 0.0061 * cos(2.0 * Mprimer - Fr)
                    - 0.0047 * sin(Mprimer + 3.0 * Fr)
                    - 0.0043 * E * sin(2.0 * Dr - Mr - Mprimer)
                    - 0.0040 * cos(Mprimer - 2.0 * Fr)
                    - 0.0037 * sin(2.0 * (Dr - Mprimer)) + 0.0031 * sin(Fr)
                    + 0.0030 * sin(2.0 * Dr + Mprimer)
                    - 0.0029 * cos(Mprimer + 2.0 * Fr)
                    - 0.0029 * E * sin(2.0 * Dr - Mr)
                    - 0.0027 * sin(Mprimer + Fr)
                    + 0.0024 * E * sin(Mr - Mprimer)
                    - 0.0021 * sin(Mprimer - 3.0 * Fr)
                    + 0.0019 * sin(2.0 * Mprimer + Fr)
                    + 0.0018 * cos(2.0 * (Dr - Mprimer) - Fr)
                    + 0.0018 * sin(3.0 * Fr) + 0.0017 * cos(Mprimer + 3.0 * Fr)
                    + 0.0017 * cos(2.0 * Mprimer)
                    - 0.0014 * cos(2.0 * Dr - Mprimer)
                    + 0.0013 * cos(2.0 * Dr + Mprimer + Fr)
                    + 0.0013 * cos(Mprimer) + 0.0012 * sin(3.0 * Mprimer + Fr)
                    + 0.0011 * sin(2.0 * Dr - Mprimer + Fr)
                    - 0.0011 * cos(2.0 * (Dr - Mprimer)) + 0.001 * cos(Dr + Fr)
                    + 0.0010 * E * sin(Mr + Mprimer)
                    - 0.0009 * sin(2.0 * (Dr - Fr))
                    + 0.0007 * cos(2.0 * Mprimer + Fr)
                    - 0.0007 * cos(3.0 * Mprimer + Fr))
            # Correction for the declination
            cor2 = (5.1093 * sin(Fr) + 0.2658 * cos(2.0 * Fr)
                    + 0.1448 * sin(2.0 * Dr - Fr) - 0.0322 * sin(3.0 * Fr)
                    + 0.0133 * cos(2.0 * (Dr - Fr)) + 0.0125 * cos(2.0 * Dr)
                    - 0.0124 * sin(Mprimer - Fr)
                    - 0.0101 * sin(Mprimer + 2.0 * Fr) + 0.0097 * cos(Fr)
                    - 0.0087 * E * sin(2.0 * Dr + Mr - Fr)
                    + 0.0074 * sin(Mprimer + 3.0 * Fr) + 0.0067 * sin(Dr + Fr)
                    + 0.0063 * sin(Mprimer - 2.0 * Fr)
                    + 0.0060 * E * sin(2.0 * Dr - Mr - Fr)
                    - 0.0057 * sin(2.0 * Dr - Mprimer - Fr)
                    - 0.0056 * cos(Mprimer + Fr)
                    + 0.0052 * cos(Mprimer + 2.0 * Fr)
                    + 0.0041 * cos(2.0 * Mprimer + Fr)
                    - 0.0040 * cos(Mprimer - 3.0 * Fr)
                    + 0.0038 * cos(2.0 * Mprimer - Fr)
                    - 0.0034 * cos(Mprimer - 2.0 * Fr)
                    - 0.0029 * sin(2.0 * Mprimer)
                    + 0.0029 * sin(3.0 * Mprimer + Fr)
                    - 0.0028 * E * cos(2.0 * Dr + Mr - Fr)
                    - 0.0028 * cos(Mprimer - Fr) - 0.0023 * cos(3.0 * Fr)
                    - 0.0021 * sin(2.0 * Dr + Fr)
                    + 0.0019 * cos(Mprimer + 3.0 * Fr) + 0.0018 * cos(Dr + Fr)
                    + 0.0017 * sin(2.0 * Mprimer - Fr)
                    + 0.0015 * cos(3.0 * Mprimer + Fr)
                    + 0.0014 * cos(2.0 * (Dr + Mprimer) + Fr)
                    - 0.0012 * sin(2.0 * (Dr - Mprimer) - Fr)
                    - 0.0012 * cos(2.0 * Mprimer) - 0.0010 * cos(Mprimer)
                    - 0.0010 * sin(2.0 * Fr) + 0.0006 * sin(Mprimer + Fr))
        else:
            # Correction for the epoch
            corr = (-0.8975 * cos(Fr) - 0.4726 * sin(Mprimer)
                    - 0.1030 * sin(2.0 * Fr) - 0.0976 * sin(2.0 * Dr - Mprimer)
                    + 0.0541 * cos(Mprimer - Fr) + 0.0516 * cos(Mprimer + Fr)
                    - 0.0438 * sin(2.0 * Dr) + 0.0112 * E * sin(Mr)
                    + 0.0157 * cos(3.0 * Fr) + 0.0023 * sin(Mprimer + 2.0 * Fr)
                    - 0.0136 * cos(2.0 * Dr - Fr)
                    + 0.0110 * cos(2.0 * Dr - Mprimer - Fr)
                    + 0.0091 * cos(2.0 * Dr - Mprimer + Fr)
                    + 0.0089 * cos(2.0 * Dr + Fr) + 0.0075 * sin(2.0 * Mprimer)
                    - 0.0030 * sin(Mprimer - 2.0 * Fr)
                    - 0.0061 * cos(2.0 * Mprimer - Fr)
                    - 0.0047 * sin(Mprimer + 3.0 * Fr)
                    - 0.0043 * E * sin(2.0 * Dr - Mr - Mprimer)
                    + 0.0040 * cos(Mprimer - 2.0 * Fr)
                    - 0.0037 * sin(2.0 * (Dr - Mprimer)) - 0.0031 * sin(Fr)
                    + 0.0030 * sin(2.0 * Dr + Mprimer)
                    + 0.0029 * cos(Mprimer + 2.0 * Fr)
                    - 0.0029 * E * sin(2.0 * Dr - Mr)
                    - 0.0027 * sin(Mprimer + Fr)
                    + 0.0024 * E * sin(Mr - Mprimer)
                    - 0.0021 * sin(Mprimer - 3.0 * Fr)
                    - 0.0019 * sin(2.0 * Mprimer + Fr)
                    - 0.0006 * cos(2.0 * (Dr - Mprimer) - Fr)
                    - 0.0018 * sin(3.0 * Fr) - 0.0017 * cos(Mprimer + 3.0 * Fr)
                    + 0.0017 * cos(2.0 * Mprimer)
                    + 0.0014 * cos(2.0 * Dr - Mprimer)
                    - 0.0013 * cos(2.0 * Dr + Mprimer + Fr)
                    - 0.0013 * cos(Mprimer) + 0.0012 * sin(3.0 * Mprimer + Fr)
                    + 0.0011 * sin(2.0 * Dr - Mprimer + Fr)
                    + 0.0011 * cos(2.0 * (Dr - Mprimer)) + 0.001 * cos(Dr + Fr)
                    + 0.0010 * E * sin(Mr + Mprimer)
                    - 0.0009 * sin(2.0 * (Dr - Fr))
                    - 0.0007 * cos(2.0 * Mprimer + Fr)
                    - 0.0007 * cos(3.0 * Mprimer + Fr))
            # Correction for the declination
            cor2 = (-5.1093 * sin(Fr) + 0.2658 * cos(2.0 * Fr)
                    - 0.1448 * sin(2.0 * Dr - Fr) + 0.0322 * sin(3.0 * Fr)
                    + 0.0133 * cos(2.0 * (Dr - Fr)) + 0.0125 * cos(2.0 * Dr)
                    - 0.0015 * sin(Mprimer - Fr)
                    + 0.0101 * sin(Mprimer + 2.0 * Fr) - 0.0097 * cos(Fr)
                    + 0.0087 * E * sin(2.0 * Dr + Mr - Fr)
                    + 0.0074 * sin(Mprimer + 3.0 * Fr) + 0.0067 * sin(Dr + Fr)
                    - 0.0063 * sin(Mprimer - 2.0 * Fr)
                    - 0.0060 * E * sin(2.0 * Dr - Mr - Fr)
                    + 0.0057 * sin(2.0 * Dr - Mprimer - Fr)
                    - 0.0056 * cos(Mprimer + Fr)
                    - 0.0052 * cos(Mprimer + 2.0 * Fr)
                    - 0.0041 * cos(2.0 * Mprimer + Fr)
                    - 0.0040 * cos(Mprimer - 3.0 * Fr)
                    - 0.0038 * cos(2.0 * Mprimer - Fr)
                    + 0.0034 * cos(Mprimer - 2.0 * Fr)
                    - 0.0029 * sin(2.0 * Mprimer)
                    + 0.0029 * sin(3.0 * Mprimer + Fr)
                    + 0.0028 * E * cos(2.0 * Dr + Mr - Fr)
                    - 0.0028 * cos(Mprimer - Fr) + 0.0023 * cos(3.0 * Fr)
                    + 0.0021 * sin(2.0 * Dr + Fr)
                    + 0.0019 * cos(Mprimer + 3.0 * Fr) + 0.0018 * cos(Dr + Fr)
                    - 0.0017 * sin(2.0 * Mprimer - Fr)
                    + 0.0015 * cos(3.0 * Mprimer + Fr)
                    + 0.0014 * cos(2.0 * (Dr + Mprimer) + Fr)
                    + 0.0012 * sin(2.0 * (Dr - Mprimer) - Fr)
                    - 0.0012 * cos(2.0 * Mprimer) + 0.0010 * cos(Mprimer)
                    - 0.0010 * sin(2.0 * Fr) + 0.0037 * sin(Mprimer + Fr))
        # Add the correction to 'jde'
        jde += corr
        jde = Epoch(jde)
        declination = 23.6961 - 0.013004 * t + cor2
        if (target == 'southern'):
            declination *= -1.0
        declination = Angle(Angle.reduce_deg(declination))
        return jde, declination

    @staticmethod
    def moon_librations(epoch):
        """This method computes the librations in longitude and latitude of the
        moon. There are several librations: The optical librations, that are
        the apparent oscillations in the hemisphere that the Moon turns towards
        the Earth, due to variations in the geometric position of the Earth
        relative to the lunar surface during the course of the orbital motion
        of the Moon. These variations allow to observe about 59% of the surface
        of the Moon from the Earth.

        There is also the physical libration of the Moon, i.e., the libration
        due to the actual rotational motion of the Moon about its mean
        rotation. The physical libration is much smaller than the optical
        libration, and can never be larger than 0.04 degree in both longitude
        and latitude.

        Finally, there is the total libration, which is the sum of the two
        librations mentioned above

        :param epoch: Epoch we want to compute the Moon's librations.
        :type year: :py:class:`Epoch`

        :returns: A tuple containing the optical libration in longitude and in
            latitude, the physical libration also in longitude and latitude,
            and the total librations which is the sum of the previouse ones,
            all of them as :py:class:`Angle` objects.
        :rtype: tuple
        :raises: TypeError if input value is of wrong type.

        >>> epoch = Epoch(1992, 4, 12.0)
        >>> lopt, bopt, lphys, bphys, ltot, btot = Moon.moon_librations(epoch)
        >>> print(round(lopt, 3))
        -1.206
        >>> print(round(bopt, 3))
        4.194
        >>> print(round(lphys, 3))
        -0.025
        >>> print(round(bphys, 3))
        0.006
        >>> print(round(ltot, 2))
        -1.23
        >>> print(round(btot, 3))
        4.2
        """

        # First check that input value is of correct type
        if not isinstance(epoch, Epoch):
            raise TypeError("Invalid input types")
        # Let's start computing some constants
        ir = Angle(1.54242).rad()
        sinI = sin(ir)
        cosI = cos(ir)
        # Now, let's call the method apparent_ecliptical_pos()
        Lambda, Beta, Delta, ppi = Moon.apparent_ecliptical_pos(epoch)
        # Compute the nutation in longitude (deltaPsi)
        deltaPsi = nutation_longitude(epoch)
        # Get the time from J2000.0 in Julian centuries
        t = (epoch - JDE2000) / 36525.0
        # Mean elongation of the Moon
        D = 297.8501921 + (445267.1114034
                           + (-0.0018819
                              + (1.0/545868.0 - t/113065000.0) * t) * t) * t
        # Sun's mean anomaly
        M = 357.5291092 + (35999.0502909 + (-0.0001536 + t/24490000.0) * t) * t
        # Moon's mean anomaly
        Mprime = 134.9633964 + (477198.8675055
                                + (0.0087414
                                   + (1.0/69699.9
                                      + t/14712000.0) * t) * t) * t
        # Moon's argument of latitude
        F = 93.2720950 + (483202.0175233
                          + (-0.0036539
                             + (-1.0/3526000.0 + t/863310000.0) * t) * t) * t
        F = Angle(Angle.reduce_deg(F)).to_positive()
        # Compute the mean longitude of the ascending node of lunar orbit
        Omega = Moon.longitude_mean_ascending_node(epoch)
        # Let's compute some additional arguments
        k1 = 119.75 + 131.849 * t
        k2 = 72.56 + 20.186 * t
        # Eccentricity of Earth's orbit around the Sun
        E = 1.0 + (-0.002516 - 0.0000074 * t) * t
        # Reduce the angles to a [0 360] range
        D = Angle(Angle.reduce_deg(D)).to_positive()
        Dr = D.rad()
        M = Angle(Angle.reduce_deg(M)).to_positive()
        Mr = M.rad()
        Mprime = Angle(Angle.reduce_deg(Mprime)).to_positive()
        Mprimer = Mprime.rad()
        F = Angle(Angle.reduce_deg(F)).to_positive()
        Fr = F.rad()
        Omegar = Omega.rad()
        k1 = Angle(Angle.reduce_deg(k1)).to_positive()
        k1r = k1.rad()
        k2 = Angle(Angle.reduce_deg(k2)).to_positive()
        k2r = k2.rad()
        # Let's compute 'w' and some additional parameters
        w = Lambda - deltaPsi - Omega
        w = w.to_positive()
        wr = w.rad()
        sinW = sin(wr)
        cosW = cos(wr)
        betar = Beta.rad()
        sinB = sin(betar)
        cosB = cos(betar)
        # Compute 'A'
        Ar = atan2((sinW * cosB * cosI - sinB * sinI), (cosW * cosB))
        A = Angle(Ar, radians=True).to_positive()
        lprime = A - F
        bprimer = asin(-sinW * cosB * sinI - sinB * cosI)
        bprime = Angle(bprimer, radians=True)
        # Compute the expressions from D.H. Eckhardt 1981
        rho = (-0.02752 * cos(Mprimer) - 0.02245 * sin(Fr)
               + 0.00684 * cos(Mprimer - 2.0 * Fr) - 0.00293 * cos(2.0 * Fr)
               - 0.00085 * cos(2.0 * (Fr - Dr))
               - 0.00054 * cos(Mprimer - 2.0 * Dr) - 0.0002 * sin(Mprimer + Fr)
               - 0.0002 * cos(Mprimer + 2.0 * Fr) - 0.0002 * cos(Mprimer - Fr)
               + 0.00014 * cos(Mprimer + 2.0 * (Fr - Dr)))
        rho = Angle(rho)
        sigma = (-0.02816 * sin(Mprimer) + 0.02244 * cos(Fr)
                 - 0.00682 * sin(Mprimer - 2.0 * Fr) - 0.00279 * sin(2.0 * Fr)
                 - 0.00083 * sin(2.0 * (Fr - Dr))
                 + 0.00069 * sin(Mprimer - 2.0 * Dr)
                 + 0.0004 * cos(Mprimer + Fr) - 0.00025 * sin(2.0 * Mprimer)
                 - 0.00023 * sin(Mprimer + 2.0 * Fr)
                 + 0.0002 * cos(Mprimer - Fr) + 0.00019 * sin(Mprimer - Fr)
                 + 0.00013 * sin(Mprimer + 2.0 * (Fr - Dr))
                 - 0.0001 * cos(Mprimer - 3.0 * Fr))
        sigma = Angle(sigma)
        tau = (0.0252 * E * sin(Mr) + 0.00473 * sin(2.0 * (Mprimer - Fr))
               - 0.00467 * sin(Mprimer) + 0.00396 * sin(k1r)
               + 0.00276 * sin(2.0 * (Mprimer - Dr)) + 0.00196 * sin(Omegar)
               - 0.00183 * cos(Mprimer - Fr)
               + 0.00115 * sin(Mprimer - 2.0 * Dr)
               - 0.00096 * sin(Mprimer - Dr) + 0.00046 * sin(2.0 * (Fr - Dr))
               - 0.00039 * sin(Mprimer - Fr) - 0.00032 * sin(Mprimer - Mr - Dr)
               + 0.00027 * sin(2.0 * (Mprimer - Dr) - Mr) + 0.00023 * sin(k2r)
               - 0.00014 * sin(2.0 * Dr) + 0.00014 * cos(2.0 * (Mprimer - Fr))
               - 0.00012 * sin(Mprimer - 2.0 * Fr)
               - 0.00012 * sin(2.0 * Mprimer)
               + 0.00011 * sin(2.0 * (Mprimer - Mr - Dr)))
        tau = Angle(tau)
        # Compute the physical librations
        lpp = -tau + (rho * cos(Ar) + sigma * sin(Ar)) * tan(bprimer)
        bpp = sigma * cos(Ar) - rho * sin(Ar)
        lt = lprime + lpp
        bt = bprime + bpp
        return lprime, bprime, lpp, bpp, lt, bt

    @staticmethod
    def moon_position_angle_axis(epoch):
        """This method computes the position angle of the Moon's axis of
        rotation. The effect of the physical libration is taken into account.

        :param epoch: Epoch we want to compute the position angle of the Moon's
            axis of rotation.
        :type year: :py:class:`Epoch`

        :returns: The position angle of the Moon's axis of rotation, as a
            :py:class:`Angle` object.
        :rtype: tuple
        :raises: TypeError if input value is of wrong type.

        >>> epoch = Epoch(1992, 4, 12.0)
        >>> p = Moon.moon_position_angle_axis(epoch)
        >>> print(round(p, 2))
        15.08
        """

        # First check that input value is of correct type
        if not isinstance(epoch, Epoch):
            raise TypeError("Invalid input types")
        # Let's start computing some constants
        ir = Angle(1.54242).rad()
        sinI = sin(ir)
        # Compute the nutation in longitude (deltaPsi)
        deltaPsi = nutation_longitude(epoch)
        # Get the true obliquity of the ecliptic
        epsilon = true_obliquity(epoch)
        epsr = epsilon.rad()
        # Get the time from J2000.0 in Julian centuries
        t = (epoch - JDE2000) / 36525.0
        # Mean elongation of the Moon
        D = 297.8501921 + (445267.1114034
                           + (-0.0018819
                              + (1.0/545868.0 - t/113065000.0) * t) * t) * t
        # Moon's mean anomaly
        Mprime = 134.9633964 + (477198.8675055
                                + (0.0087414
                                   + (1.0/69699.9
                                      + t/14712000.0) * t) * t) * t
        # Moon's argument of latitude
        F = 93.2720950 + (483202.0175233
                          + (-0.0036539
                             + (-1.0/3526000.0 + t/863310000.0) * t) * t) * t
        F = Angle(Angle.reduce_deg(F)).to_positive()
        # Compute the mean longitude of the ascending node of lunar orbit
        Omega = Moon.longitude_mean_ascending_node(epoch)
        # Reduce the angles to a [0 360] range
        D = Angle(Angle.reduce_deg(D)).to_positive()
        Dr = D.rad()
        Mprime = Angle(Angle.reduce_deg(Mprime)).to_positive()
        Mprimer = Mprime.rad()
        F = Angle(Angle.reduce_deg(F)).to_positive()
        Fr = F.rad()
        # Compute the expressions from D.H. Eckhardt 1981
        rho = (-0.02752 * cos(Mprimer) - 0.02245 * sin(Fr)
               + 0.00684 * cos(Mprimer - 2.0 * Fr) - 0.00293 * cos(2.0 * Fr)
               - 0.00085 * cos(2.0 * (Fr - Dr))
               - 0.00054 * cos(Mprimer - 2.0 * Dr) - 0.0002 * sin(Mprimer + Fr)
               - 0.0002 * cos(Mprimer + 2.0 * Fr) - 0.0002 * cos(Mprimer - Fr)
               + 0.00014 * cos(Mprimer + 2.0 * (Fr - Dr)))
        rho = Angle(rho)
        rhor = rho.rad()
        sigma = (-0.02816 * sin(Mprimer) + 0.02244 * cos(Fr)
                 - 0.00682 * sin(Mprimer - 2.0 * Fr) - 0.00279 * sin(2.0 * Fr)
                 - 0.00083 * sin(2.0 * (Fr - Dr))
                 + 0.00069 * sin(Mprimer - 2.0 * Dr)
                 + 0.0004 * cos(Mprimer + Fr) - 0.00025 * sin(2.0 * Mprimer)
                 - 0.00023 * sin(Mprimer + 2.0 * Fr)
                 + 0.0002 * cos(Mprimer - Fr) + 0.00019 * sin(Mprimer - Fr)
                 + 0.00013 * sin(Mprimer + 2.0 * (Fr - Dr))
                 - 0.0001 * cos(Mprimer - 3.0 * Fr))
        sigma = Angle(sigma)
        # Compute the parameters 'v', 'x', 'y' and 'w'
        v = Omega + deltaPsi + (sigma / sinI)
        vr = v.rad()
        x = sin(ir + rhor) * sin(vr)
        y = sin(ir + rhor) * cos(vr) * cos(epsr) - cos(ir + rhor) * sin(epsr)
        w = atan2(x, y)
        # Now, let's call the method apparent_equatorial_pos()
        alpha, dec, Delta, ppi = Moon.apparent_equatorial_pos(epoch)
        alphar = alpha.rad()
        # Get the Moon librations
        lopt, bopt, lphys, bphys, ltot, btot = Moon.moon_librations(epoch)
        p = asin((sqrt(x * x + y * y) * cos(alphar - w)) / cos(btot.rad()))
        return Angle(p, radians=True)


def main():

    # Let's define a small helper function
    def print_me(msg, val):
        print("{}: {}".format(msg, val))

    # Let's show some uses of Saturn class
    print("\n" + 35 * "*")
    print("*** Use of Moon class")
    print(35 * "*" + "\n")

    # Let's compute the Moon geocentric ecliiptical position for a given epoch
    epoch = Epoch(1992, 4, 12.0)
    Lambda, Beta, Delta, ppi = Moon.geocentric_ecliptical_pos(epoch)
    print_me("Longitude (Lambda)", round(Lambda, 6))    # 133.162655
    print_me("Latitude (Beta)", round(Beta, 6))         # -3.229126
    print_me("Distance (Delta)", round(Delta, 1))       # 368409.7
    print_me("Equatorial horizontal parallax (Pi)", round(ppi, 6))  # 0.991990

    print("")

    # Now let's compute the apparent ecliptical position
    epoch = Epoch(1992, 4, 12.0)
    Lambda, Beta, Delta, ppi = Moon.apparent_ecliptical_pos(epoch)
    print_me("Longitude (Lambda)", round(Lambda, 6))    # 133.167264
    print_me("Latitude (Beta)", round(Beta, 6))         # -3.229126
    print_me("Distance (Delta)", round(Delta, 1))       # 368409.7
    print_me("Equatorial horizontal parallax (Pi)", round(ppi, 6))  # 0.991990

    print("")

    # Get the apparent equatorial position
    epoch = Epoch(1992, 4, 12.0)
    ra, dec, Delta, ppi = Moon.apparent_equatorial_pos(epoch)
    print_me("Right Ascension (ra)", round(ra, 6))      # 134.688469
    print_me("Declination (dec)", round(dec, 6))        # 13.768367
    print_me("Distance (Delta)", round(Delta, 1))       # 368409.7
    print_me("Equatorial horizontal parallax (Pi)", round(ppi, 6))  # 0.991990

    print("")

    # Compute the longitude of the Moon's mean ascending node
    epoch = Epoch(1913, 5, 27.0)
    Omega = Moon.longitude_mean_ascending_node(epoch)
    print_me("Longitude of the mean ascending node", round(Omega, 1))   # 0.0
    epoch = Epoch(1959, 12, 7.0)
    Omega = Moon.longitude_mean_ascending_node(epoch)
    print_me("Longitude of the mean ascending node", round(Omega, 1))   # 180.0

    print("")

    # Get the longitude of the Moonś true ascending node
    epoch = Epoch(1913, 5, 27.0)
    Omega = Moon.longitude_true_ascending_node(epoch)
    print_me("Longitude of the true ascending node", round(Omega, 4))  # 0.8763

    print("")

    # Compute the longitude of the Moon's mean perigee
    epoch = Epoch(2021, 3, 5.0)
    Pi = Moon.longitude_mean_perigee(epoch)
    print_me("Longitude of the mean perigee", round(Pi, 5))     # 224.89194

    print("")

    # Compute the approximate illuminated fraction of the Moon's disk
    epoch = Epoch(1992, 4, 12.0)
    k = Moon.illuminated_fraction_disk(epoch)
    print_me("Approximate illuminated fraction of Moon's disk", round(k, 2))
    # 0.68

    print("")

    # Compute the position angle of the bright limb of the Moon
    epoch = Epoch(1992, 4, 12.0)
    xi = Moon.position_bright_limb(epoch)
    print_me("Position angle of the bright limb of the Moon", round(xi, 1))
    # 285.0

    print("")

    # Calculate the instant of a New Moon
    epoch = Epoch(1977, 2, 15.0)
    new_moon = Moon.moon_phase(epoch, target="new")
    y, m, d, h, mi, s = new_moon.get_full_date()
    print("New Moon: {}/{}/{} {}:{}:{}".format(y, m, d, h, mi, round(s)))
    # 1977/2/18 3:37:42

    # Calculate the time of a Last Quarter
    epoch = Epoch(2044, 1, 1.0)
    new_moon = Moon.moon_phase(epoch, target="last")
    y, m, d, h, mi, s = new_moon.get_full_date()
    print("Last Quarter: {}/{}/{} {}:{}:{}".format(y, m, d, h, mi, round(s)))
    # 2044/1/21 23:48:17

    print("")

    # Compute the time and parallax of apogee
    epoch = Epoch(1988, 10, 1.0)
    apogee, parallax = Moon.moon_perigee_apogee(epoch, target="apogee")
    y, m, d, h, mi, s = apogee.get_full_date()
    print("Apogee epoch: {}/{}/{} {}:{}".format(y, m, d, h, mi))
    # 1988/10/7 20:30
    print_me("Equatorial horizontal parallax", parallax.dms_str(n_dec=3))
    # 54' 0.679''

    print("")

    # Compute the time of passage by the ascending node
    epoch = Epoch(1987, 5, 15.0)
    passage = Moon.moon_passage_nodes(epoch, target="ascending")
    y, m, d, h, mi, s = passage.get_full_date()
    mi += s/60.0
    print("Passage by the ascending node: {}/{}/{} {}:{}".format(y,
                                                                 m,
                                                                 d,
                                                                 h,
                                                                 round(mi)))
    # 1987/5/23 6:26

    print("")

    # Compute the epoch and amplitude of maximum southern declination
    epoch = Epoch(2049, 4, 15.0)
    epo, dec = Moon.moon_maximum_declination(epoch, target='southern')
    y, m, d, h, mi, s = epo.get_full_date()
    print("Epoch of maximum declination: {}/{}/{} {}:{}".format(y, m, d, h,
                                                                mi))
    # 2049/4/21 14:0
    print_me("Amplitude of maximum declination", dec.dms_str(n_dec=0))
    # -22d 8' 18.0''

    print("")

    # Compute the librations of the Moon
    epoch = Epoch(1992, 4, 12.0)
    lopt, bopt, lphys, bphys, ltot, btot = Moon.moon_librations(epoch)
    print_me("Optical libration in longitude", round(lopt, 3))
    # -1.206
    print_me("Optical libration in latitude", round(bopt, 3))
    # 4.194
    print_me("Physical libration in longitude", round(lphys, 3))
    # -0.025
    print_me("Physical libration in latitude", round(bphys, 3))
    # 0.006
    print_me("Total libration in longitude", round(lphys, 2))
    # -1.23
    print_me("Total libration in latitude", round(bphys, 3))
    # 4.2

    print("")

    # Let's calculate the position angle of the Moon's axis of rotation
    epoch = Epoch(1992, 4, 12.0)
    p = Moon.moon_position_angle_axis(epoch)
    print_me("Position angle of Moon's axis of rotation", round(p, 2))
    # 15.08


if __name__ == "__main__":

    main()



# encoding: utf-8

"""
Custom element classes related to the styles part
"""

from ..enum.style import WD_STYLE_TYPE
from .simpletypes import ST_DecimalNumber, ST_OnOff, ST_String
from .xmlchemy import (
    BaseOxmlElement, OptionalAttribute, RequiredAttribute, ZeroOrMore,
    ZeroOrOne
)


def styleId_from_name(name):
    """
    Return the style id corresponding to *name*, taking into account
    special-case names such as 'Heading 1'.
    """
    return {
        'caption':   'Caption',
        'heading 1': 'Heading1',
        'heading 2': 'Heading2',
        'heading 3': 'Heading3',
        'heading 4': 'Heading4',
        'heading 5': 'Heading5',
        'heading 6': 'Heading6',
        'heading 7': 'Heading7',
        'heading 8': 'Heading8',
        'heading 9': 'Heading9',
    }.get(name, name.replace(' ', ''))


class CT_LatentStyles(BaseOxmlElement):
    """
    `w:latentStyles` element, defining behavior defaults for latent styles
    and containing `w:lsdException` child elements that each override those
    defaults for a named latent style.
    """
    lsdException = ZeroOrMore('w:lsdException', successors=())

    count = OptionalAttribute('w:count', ST_DecimalNumber)
    defLockedState = OptionalAttribute('w:defLockedState', ST_OnOff)
    defQFormat = OptionalAttribute('w:defQFormat', ST_OnOff)
    defSemiHidden = OptionalAttribute('w:defSemiHidden', ST_OnOff)
    defUIPriority = OptionalAttribute('w:defUIPriority', ST_DecimalNumber)
    defUnhideWhenUsed = OptionalAttribute('w:defUnhideWhenUsed', ST_OnOff)

    def bool_prop(self, attr_name):
        """
        Return the boolean value of the attribute having *attr_name*, or
        |False| if not present.
        """
        value = getattr(self, attr_name)
        if value is None:
            return False
        return value

    def get_by_name(self, name):
        """
        Return the `w:lsdException` child having *name*, or |None| if not
        found.
        """
        found = self.xpath('w:lsdException[@w:name="%s"]' % name)
        if not found:
            return None
        return found[0]

    def set_bool_prop(self, attr_name, value):
        """
        Set the on/off attribute having *attr_name* to *value*.
        """
        setattr(self, attr_name, bool(value))


class CT_LsdException(BaseOxmlElement):
    """
    ``<w:lsdException>`` element, defining override visibility behaviors for
    a named latent style.
    """
    locked = OptionalAttribute('w:locked', ST_OnOff)
    name = RequiredAttribute('w:name', ST_String)
    qFormat = OptionalAttribute('w:qFormat', ST_OnOff)
    semiHidden = OptionalAttribute('w:semiHidden', ST_OnOff)
    uiPriority = OptionalAttribute('w:uiPriority', ST_DecimalNumber)
    unhideWhenUsed = OptionalAttribute('w:unhideWhenUsed', ST_OnOff)

    def delete(self):
        """
        Remove this `w:lsdException` element from the XML document.
        """
        self.getparent().remove(self)

    def on_off_prop(self, attr_name):
        """
        Return the boolean value of the attribute having *attr_name*, or
        |None| if not present.
        """
        return getattr(self, attr_name)

    def set_on_off_prop(self, attr_name, value):
        """
        Set the on/off attribute having *attr_name* to *value*.
        """
        setattr(self, attr_name, value)


class CT_Style(BaseOxmlElement):
    """
    A ``<w:style>`` element, representing a style definition
    """
    _tag_seq = (
        'w:name', 'w:aliases', 'w:basedOn', 'w:next', 'w:link',
        'w:autoRedefine', 'w:hidden', 'w:uiPriority', 'w:semiHidden',
        'w:unhideWhenUsed', 'w:qFormat', 'w:locked', 'w:personal',
        'w:personalCompose', 'w:personalReply', 'w:rsid', 'w:pPr', 'w:rPr',
        'w:tblPr', 'w:trPr', 'w:tcPr', 'w:tblStylePr'
    )
    name = ZeroOrOne('w:name', successors=_tag_seq[1:])
    basedOn = ZeroOrOne('w:basedOn', successors=_tag_seq[3:])
    next = ZeroOrOne('w:next', successors=_tag_seq[4:])
    uiPriority = ZeroOrOne('w:uiPriority', successors=_tag_seq[8:])
    semiHidden = ZeroOrOne('w:semiHidden', successors=_tag_seq[9:])
    unhideWhenUsed = ZeroOrOne('w:unhideWhenUsed', successors=_tag_seq[10:])
    qFormat = ZeroOrOne('w:qFormat', successors=_tag_seq[11:])
    locked = ZeroOrOne('w:locked', successors=_tag_seq[12:])
    pPr = ZeroOrOne('w:pPr', successors=_tag_seq[17:])
    rPr = ZeroOrOne('w:rPr', successors=_tag_seq[18:])
    del _tag_seq

    type = OptionalAttribute('w:type', WD_STYLE_TYPE)
    styleId = OptionalAttribute('w:styleId', ST_String)
    default = OptionalAttribute('w:default', ST_OnOff)
    customStyle = OptionalAttribute('w:customStyle', ST_OnOff)

    @property
    def basedOn_val(self):
        """
        Value of `w:basedOn/@w:val` or |None| if not present.
        """
        basedOn = self.basedOn
        if basedOn is None:
            return None
        return basedOn.val

    @basedOn_val.setter
    def basedOn_val(self, value):
        if value is None:
            self._remove_basedOn()
        else:
            self.get_or_add_basedOn().val = value

    @property
    def base_style(self):
        """
        Sibling CT_Style element this style is based on or |None| if no base
        style or base style not found.
        """
        basedOn = self.basedOn
        if basedOn is None:
            return None
        styles = self.getparent()
        base_style = styles.get_by_id(basedOn.val)
        if base_style is None:
            return None
        return base_style

    def delete(self):
        """
        Remove this `w:style` element from its parent `w:styles` element.
        """
        self.getparent().remove(self)

    @property
    def locked_val(self):
        """
        Value of `w:locked/@w:val` or |False| if not present.
        """
        locked = self.locked
        if locked is None:
            return False
        return locked.val

    @locked_val.setter
    def locked_val(self, value):
        self._remove_locked()
        if bool(value) is True:
            locked = self._add_locked()
            locked.val = value

    @property
    def name_val(self):
        """
        Value of ``<w:name>`` child or |None| if not present.
        """
        name = self.name
        if name is None:
            return None
        return name.val

    @name_val.setter
    def name_val(self, value):
        self._remove_name()
        if value is not None:
            name = self._add_name()
            name.val = value

    @property
    def next_style(self):
        """
        Sibling CT_Style element identified by the value of `w:name/@w:val`
        or |None| if no value is present or no style with that style id
        is found.
        """
        next = self.next
        if next is None:
            return None
        styles = self.getparent()
        return styles.get_by_id(next.val)  # None if not found

    @property
    def qFormat_val(self):
        """
        Value of `w:qFormat/@w:val` or |False| if not present.
        """
        qFormat = self.qFormat
        if qFormat is None:
            return False
        return qFormat.val

    @qFormat_val.setter
    def qFormat_val(self, value):
        self._remove_qFormat()
        if bool(value):
            self._add_qFormat()

    @property
    def semiHidden_val(self):
        """
        Value of ``<w:semiHidden>`` child or |False| if not present.
        """
        semiHidden = self.semiHidden
        if semiHidden is None:
            return False
        return semiHidden.val

    @semiHidden_val.setter
    def semiHidden_val(self, value):
        self._remove_semiHidden()
        if bool(value) is True:
            semiHidden = self._add_semiHidden()
            semiHidden.val = value

    @property
    def uiPriority_val(self):
        """
        Value of ``<w:uiPriority>`` child or |None| if not present.
        """
        uiPriority = self.uiPriority
        if uiPriority is None:
            return None
        return uiPriority.val

    @uiPriority_val.setter
    def uiPriority_val(self, value):
        self._remove_uiPriority()
        if value is not None:
            uiPriority = self._add_uiPriority()
            uiPriority.val = value

    @property
    def unhideWhenUsed_val(self):
        """
        Value of `w:unhideWhenUsed/@w:val` or |False| if not present.
        """
        unhideWhenUsed = self.unhideWhenUsed
        if unhideWhenUsed is None:
            return False
        return unhideWhenUsed.val

    @unhideWhenUsed_val.setter
    def unhideWhenUsed_val(self, value):
        self._remove_unhideWhenUsed()
        if bool(value) is True:
            unhideWhenUsed = self._add_unhideWhenUsed()
            unhideWhenUsed.val = value


class CT_Styles(BaseOxmlElement):
    """
    ``<w:styles>`` element, the root element of a styles part, i.e.
    styles.xml
    """
    _tag_seq = ('w:docDefaults', 'w:latentStyles', 'w:style')
    latentStyles = ZeroOrOne('w:latentStyles', successors=_tag_seq[2:])
    style = ZeroOrMore('w:style', successors=())
    del _tag_seq

    def add_style_of_type(self, name, style_type, builtin):
        """
        Return a newly added `w:style` element having *name* and
        *style_type*. `w:style/@customStyle` is set based on the value of
        *builtin*.
        """
        style = self.add_style()
        style.type = style_type
        style.customStyle = None if builtin else True
        style.styleId = styleId_from_name(name)
        style.name_val = name
        return style

    def default_for(self, style_type):
        """
        Return `w:style[@w:type="*{style_type}*][-1]` or |None| if not found.
        """
        default_styles_for_type = [
            s for s in self._iter_styles()
            if s.type == style_type and s.default
        ]
        if not default_styles_for_type:
            return None
        # spec calls for last default in document order
        return default_styles_for_type[-1]

    def get_by_id(self, styleId):
        """
        Return the ``<w:style>`` child element having ``styleId`` attribute
        matching *styleId*, or |None| if not found.
        """
        xpath = 'w:style[@w:styleId="%s"]' % styleId
        try:
            return self.xpath(xpath)[0]
        except IndexError:
            return None

    def get_by_name(self, name):
        """
        Return the ``<w:style>`` child element having ``<w:name>`` child
        element with value *name*, or |None| if not found.
        """
        xpath = 'w:style[w:name/@w:val="%s"]' % name
        try:
            return self.xpath(xpath)[0]
        except IndexError:
            return None

    def _iter_styles(self):
        """
        Generate each of the `w:style` child elements in document order.
        """
        return (style for style in self.xpath('w:style'))



# -*- coding: utf-8 -*-


# PyMeeus: Python module implementing astronomical algorithms.
# Copyright (C) 2018  Dagoberto Salazar
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import sqrt, radians, fsum, sin

from pymeeus.base import TOL
from pymeeus.Angle import Angle


"""
.. module:: CurveFitting
   :synopsis: Class to get the best fit of a curve to a set of (x, y) points
   :license: GNU Lesser General Public License v3 (LGPLv3)

.. moduleauthor:: Dagoberto Salazar
"""


class CurveFitting(object):
    """
    Class CurveFitting deals with finding the function (linear, cuadratic, etc)
    that best fit a given set of points.

    The constructor takes pairs of (x, y) values from the table of interest.
    These pairs of values can be given as a sequence of int/floats, tuples,
    lists or Angles. It is also possible to provide a CurveFitting object to
    the constructor in order to get a copy.

    .. note:: When using Angles, be careful with the 360-to-0 discontinuity.

    If a sequence of int, floats or Angles is given, the values in the odd
    positions are considered to belong to the 'x' set, while the values in the
    even positions belong to the 'y' set. If only one tuple or list is
    provided, it is assumed that it is the 'y' set, and the 'x' set is build
    from 0 onwards with steps of length 1.

    Please keep in mind that a minimum of two data pairs are needed in order to
    carry out any fitting. If only one value pair is provided, a ValueError
    exception will be raised.
    """

    def __init__(self, *args):
        """CurveFitting constructor.

        This takes pairs of (x, y) values from the table of interest. These
        pairs of values can be given as a sequence of int/floats, tuples, lists
        or Angles. It is also possible to provide a CurveFitting object to the
        constructor in order to get a copy.

        .. note:: When using Angles, be careful with the 360-to-0 discontinuity

        If a sequence of int, floats or Angles is given, the values in the odd
        positions are considered to belong to the 'x' set, while the values in
        the even positions belong to the 'y' set. If only one tuple or list is
        provided, it is assumed that it is the 'y' set, and the 'x' set is
        build from 0 onwards with steps of length 1.

        Please keep in mind that a minimum of two data pairs are needed in
        order to carry out any interpolation. If only one value pair is
        provided, a ValueError exception will be raised.

        :param args: Input tabular values, or another CurveFitting object.
        :type args: int, float, list, tuple, :py:class:`Angle`,
           :py:class:`CurveFitting`

        :returns: CurveFitting object.
        :rtype: :py:class:`CurveFitting`
        :raises: ValueError if not enough input data pairs are provided.
        :raises: TypeError if input values are of wrong type.

        >>> i = CurveFitting([5, 3, 6, 1, 2, 4, 9], [10, 6, 12, 2, 4, 8])
        >>> print(i._x)
        [5, 3, 6, 1, 2, 4]
        >>> print(i._y)
        [10, 6, 12, 2, 4, 8]
        >>> j = CurveFitting([3, -8, 1, 12, 2, 5, 8])
        >>> print(j._x)
        [0, 1, 2, 3, 4, 5, 6]
        >>> print(j._y)
        [3, -8, 1, 12, 2, 5, 8]
        >>> k = CurveFitting(3, -8, 1, 12, 2, 5, 8)
        >>> print(k._x)
        [3, 1, 2]
        >>> print(k._y)
        [-8, 12, 5]
        >>> m = CurveFitting(k)
        >>> print(m._x)
        [3, 1, 2]
        >>> print(m._y)
        [-8, 12, 5]
        """

        # Initialize data table
        self._x = []
        self._y = []
        self.set(*args)  # Let's use 'set()' method to handle the setup

    def set(self, *args):
        """Method used to define the value pairs of CurveFitting object.

        This takes pairs of (x, y) values from the table of interest. These
        pairs of values can be given as a sequence of int/floats, tuples,
        lists, or Angles. It is also possible to provide a CurveFitting object
        to this method in order to get a copy.

        .. note:: When using Angles, be careful with the 360-to-0 discontinuity

        If a sequence of int, floats or Angles is given, the values in the odd
        positions are considered to belong to the 'x' set, while the values in
        the even positions belong to the 'y' set. If only one tuple or list is
        provided, it is assumed that it is the 'y' set, and the 'x' set is
        build from 0 onwards with steps of length 1.

        Please keep in mind that a minimum of two data pairs are needed in
        order to carry out any interpolation. If only one value is provided, a
        ValueError exception will be raised.

        :param args: Input tabular values, or another CurveFitting object.
        :type args: int, float, list, tuple, :py:class:`Angle`

        :returns: None.
        :rtype: None
        :raises: ValueError if not enough input data pairs are provided.
        :raises: TypeError if input values are of wrong type.

        >>> i = CurveFitting()
        >>> i.set([5, 3, 6, 1, 2, 4, 9], [10, 6, 12, 2, 4, 8])
        >>> print(i._x)
        [5, 3, 6, 1, 2, 4]
        >>> print(i._y)
        [10, 6, 12, 2, 4, 8]
        >>> j = CurveFitting()
        >>> j.set([3, -8, 1, 12, 2, 5, 8])
        >>> print(j._x)
        [0, 1, 2, 3, 4, 5, 6]
        >>> print(j._y)
        [3, -8, 1, 12, 2, 5, 8]
        >>> k = CurveFitting(3, -8, 1, 12, 2, 5, 8)
        >>> print(k._x)
        [3, 1, 2]
        >>> print(k._y)
        [-8, 12, 5]
        """

        # Clean up the internal data tables and parameters
        self._x = []
        self._y = []
        # If no arguments are given, return. Internal data tables are empty
        if len(args) == 0:
            return
        # If we have only one argument, it can be a single value or tuple/list
        elif len(args) == 1:
            if isinstance(args[0], CurveFitting):
                self._x = args[0]._x
                self._y = args[0]._y
            elif isinstance(args[0], (int, float, Angle)):
                # Insuficient data for curve fitting. Raise ValueError
                raise ValueError("Invalid number of input values")
            elif isinstance(args[0], (list, tuple)):
                seq = args[0]
                if len(seq) < 2:
                    raise ValueError("Invalid number of input values")
                else:
                    # Read input values into 'y', and create 'x'
                    i = 0
                    for value in seq:
                        self._x.append(i)
                        self._y.append(value)
                        i += 1
            else:
                raise TypeError("Invalid input value")
        elif len(args) == 2:
            if isinstance(args[0], (int, float, Angle)) or isinstance(
                args[1], (int, float, Angle)
            ):
                # Insuficient data for curve fitting. Raise ValueError
                raise ValueError("Invalid number of input values")
            elif isinstance(args[0], (list, tuple)) and isinstance(
                args[1], (list, tuple)
            ):
                x = args[0]
                y = args[1]
                # Check if they have the same length. If not, make them equal
                length_min = min(len(x), len(y))
                x = x[:length_min]
                y = y[:length_min]
                if len(x) < 2 or len(y) < 2:
                    raise ValueError("Invalid number of input values")
                else:
                    # Read input values into 'x' and 'y'
                    for xval, yval in zip(x, y):
                        self._x.append(xval)
                        self._y.append(yval)
            else:
                raise TypeError("Invalid input value")
        elif len(args) == 3:
            # In this case, no combination of input values is valid
            raise ValueError("Invalid number of input values")
        else:
            # If there is an odd number of arguments, drop the last one
            if len(args) % 2 != 0:
                args = args[:-1]
            # Check that all the arguments are ints, floats or Angles
            all_numbers = True
            for arg in args:
                all_numbers = (all_numbers
                               and isinstance(arg, (int, float, Angle)))
            # If any of the values failed the test, raise an exception
            if not all_numbers:
                raise TypeError("Invalid input value")
            # Now, extract the data: Odds are x's, evens are y's
            for i in range(int(len(args) / 2.0)):
                self._x.append(args[2 * i])
                self._y.append(args[2 * i + 1])
        # Compute parameters
        if len(self._x) > 0:
            self._compute_parameters()

    def _compute_parameters(self):
        """Method to compute the intermediate parameters using for fitting."""

        self._P = 0.0
        self._Q = 0.0
        self._R = 0.0
        self._S = 0.0
        self._T = 0.0
        self._U = 0.0
        self._V = 0.0
        self._W = 0.0
        self._N = len(self._x)
        self._P = fsum(self._x)
        self._T = fsum(self._y)
        for i in range(self._N):
            x2 = self._x[i] * self._x[i]
            xy = self._x[i] * self._y[i]
            self._Q += x2
            self._R += x2 * self._x[i]
            self._S += x2 * x2
            self._U += xy
            self._V += xy * self._x[i]
            self._W += self._y[i] * self._y[i]
        return

    def __str__(self):
        """Method used when trying to print the object.

        :returns: Internal tabular values as strings.
        :rtype: string

        >>> i = CurveFitting([5, 3, 6, 1, 2, 4, 9], [10, 6, 12, 2, 4, 8])
        >>> print(i)
        X: [5, 3, 6, 1, 2, 4]
        Y: [10, 6, 12, 2, 4, 8]
        """

        xstr = "X: " + str(self._x) + "\n"
        ystr = "Y: " + str(self._y)
        return xstr + ystr

    def __repr__(self):
        """Method providing the 'official' string representation of the object.
        It provides a valid expression that could be used to recreate the
        object.

        :returns: As string with a valid expression to recreate the object
        :rtype: string

        >>> i = CurveFitting([5, 3, 6, 1, 2, 4, 9], [10, 6, 12, 2, 4, 8])
        >>> repr(i)
        'CurveFitting([5, 3, 6, 1, 2, 4], [10, 6, 12, 2, 4, 8])'
        """

        return "{}({}, {})".format(self.__class__.__name__, self._x, self._y)

    def __len__(self):
        """This method returns the number of value pairs internally stored in
        this object.

        :returns: Number of value pairs internally stored
        :rtype: int

        >>> i = CurveFitting([5, 3, 6, 1, 2, 4, 9], [10, 6, 12, 2, 4, 8])
        >>> len(i)
        6
        """

        return len(self._x)

    def correlation_coeff(self):
        """This method returns the coefficient of correlation, as a float.

        :returns: Coefficient of correlation.
        :rtype: float

        >>> cf = CurveFitting([73.0, 38.0, 35.0, 42.0, 78.0, 68.0, 74.0, 42.0,
        ...                    52.0, 54.0, 39.0, 61.0, 42.0, 49.0, 50.0, 62.0,
        ...                    44.0, 39.0, 43.0, 54.0, 44.0, 37.0],
        ...                   [90.4, 125.3, 161.8, 143.4, 52.5, 50.8, 71.5,
        ...                    152.8, 131.3, 98.5, 144.8, 78.1, 89.5, 63.9,
        ...                    112.1, 82.0, 119.8, 161.2, 208.4, 111.6, 167.1,
        ...                    162.1])
        >>> r = cf.correlation_coeff()
        >>> print(round(r, 3))
        -0.767
        """

        n = self._N
        sxy = self._U
        sx = self._P
        sy = self._T
        sx2 = self._Q
        sy2 = self._W
        return ((n * sxy - sx * sy) / (sqrt(n * sx2 - sx * sx)
                                       * sqrt(n * sy2 - sy * sy)))

    def linear_fitting(self):
        """This method returns a tuple with the 'a', 'b' coefficients of the
        linear equation *'y = a*x + b'* that best fits the table data, using
        the least squares approach.

        :returns: 'a', 'b' coefficients of best linear equation fit.
        :rtype: tuple
        :raises: ZeroDivisionError if input data leads to a division by zero

        >>> cf = CurveFitting([73.0, 38.0, 35.0, 42.0, 78.0, 68.0, 74.0, 42.0,
        ...                    52.0, 54.0, 39.0, 61.0, 42.0, 49.0, 50.0, 62.0,
        ...                    44.0, 39.0, 43.0, 54.0, 44.0, 37.0],
        ...                   [90.4, 125.3, 161.8, 143.4, 52.5, 50.8, 71.5,
        ...                    152.8, 131.3, 98.5, 144.8, 78.1, 89.5, 63.9,
        ...                    112.1, 82.0, 119.8, 161.2, 208.4, 111.6, 167.1,
        ...                    162.1])
        >>> a, b = cf.linear_fitting()
        >>> print("a = {}\tb = {}".format(round(a, 2), round(b, 2)))
        a = -2.49	b = 244.18
        """

        n = self._N
        sxy = self._U
        sx = self._P
        sy = self._T
        sx2 = self._Q
        d = n * sx2 - sx * sx

        if abs(d) < TOL:
            raise ZeroDivisionError("Input data leads to a division by zero")

        a = (n * sxy - sx * sy) / d
        b = (sy * sx2 - sx * sxy) / d
        return (a, b)

    def quadratic_fitting(self):
        """This method returns a tuple with the 'a', 'b', 'c' coefficients of
        the quadratic equation *'y = a*x*x + b*x + c'* that best fits the table
        data, using the least squares approach.

        :returns: 'a', 'b', 'c' coefficients of best quadratic equation fit.
        :rtype: tuple
        :raises: ZeroDivisionError if input data leads to a division by zero

        >>> cf2 = CurveFitting([-2.0, -1.5, -1.0, -0.5, 0.0, 0.5, 1.0, 1.5,
        ...                     2.0, 2.5,3.0],
        ...                    [-9.372, -3.821, 0.291, 3.730, 5.822, 8.324,
        ...                     9.083, 6.957, 7.006, 0.365, -1.722])
        >>> a, b, c = cf2.quadratic_fitting()
        >>> print("a = {}; b = {}; c = {}".format(round(a, 2), round(b, 2),
        ...                                       round(c, 2)))
        a = -2.22; b = 3.76; c = 6.64
        """

        n = self._N
        p = self._P
        q = self._Q
        r = self._R
        s = self._S
        t = self._T
        u = self._U
        v = self._V
        q2 = q * q
        d = n * q * s + 2.0 * p * q * r - q2 * q - p * p * s - n * r * r

        if abs(d) < TOL:
            raise ZeroDivisionError("Input data leads to a division by zero")

        a = (n * q * v + p * r * t + p * q * u
             - q2 * t - p * p * v - n * r * u) / d
        b = (n * s * u + p * q * v + q * r * t
             - q2 * u - p * s * t - n * r * v) / d
        c = (q * s * t + q * r * u + p * r * v
             - q2 * v - p * s * u - r * r * t) / d
        return (a, b, c)

    def general_fitting(self, f0, f1=lambda *args: 0.0, f2=lambda *args: 0.0):
        """This method returns a tuple with the 'a', 'b', 'c' coefficients of
        the general equation *'y = a*f0(x) + b*f1(x) + c*f2(x)'* that best fits
        the table data, using the least squares approach.

        :param f0, f1, f2: Functions used to build the general equation.
        :type f0, f1, f2: function
        :returns: 'a', 'b', 'c' coefficients of best general equation fit.
        :rtype: tuple
        :raises: ZeroDivisionError if input functions are null or input data
           leads to a division by zero

        >>> cf4 = CurveFitting([3, 20, 34, 50, 75, 88, 111, 129, 143, 160, 183,
        ...                     200, 218, 230, 248, 269, 290, 303, 320, 344],
        ...                    [0.0433, 0.2532, 0.3386, 0.3560, 0.4983, 0.7577,
        ...                     1.4585, 1.8628, 1.8264, 1.2431, -0.2043,
        ...                     -1.2431, -1.8422, -1.8726, -1.4889, -0.8372,
        ...                     -0.4377, -0.3640, -0.3508, -0.2126])
        >>> def sin1(x): return sin(radians(x))
        >>> def sin2(x): return sin(radians(2.0*x))
        >>> def sin3(x): return sin(radians(3.0*x))
        >>> a, b, c = cf4.general_fitting(sin1, sin2, sin3)
        >>> print("a = {}; b = {}; c = {}".format(round(a, 2), round(b, 2),
        ...                                       round(c, 2)))
        a = 1.2; b = -0.77; c = 0.39

        >>> cf5 = CurveFitting([0, 1.2, 1.4, 1.7, 2.1, 2.2])
        >>> a, b, c = cf5.general_fitting(sqrt)
        >>> print("a = {}; b = {}; c = {}".format(round(a, 3), round(b, 3),
        ...                                       round(c, 3)))
        a = 1.016; b = 0.0; c = 0.0
        """

        m = 0
        p = 0
        q = 0
        r = 0
        s = 0
        t = 0
        u = 0
        v = 0
        w = 0
        xl = list(self._x)
        yl = list(self._y)
        for i, value in enumerate(xl):
            x = value
            y = yl[i]
            m += f0(x) * f0(x)
            p += f0(x) * f1(x)
            q += f0(x) * f2(x)
            r += f1(x) * f1(x)
            s += f1(x) * f2(x)
            t += f2(x) * f2(x)
            u += y * f0(x)
            v += y * f1(x)
            w += y * f2(x)

        if abs(r) < TOL and abs(t) < TOL and abs(m) >= TOL:
            return (u / m, 0.0, 0.0)

        if abs(m * r * t) < TOL:
            raise ZeroDivisionError("Invalid input functions: They are null")

        d = m * r * t + 2.0 * p * q * s - m * s * s - r * q * q - t * p * p

        if abs(d) < TOL:
            raise ZeroDivisionError("Input data leads to a division by zero")

        a = (u * (r * t - s * s) + v * (q * s - p * t)
             + w * (p * s - q * r)) / d
        b = (u * (s * q - p * t) + v * (m * t - q * q)
             + w * (p * q - m * s)) / d
        c = (u * (p * s - r * q) + v * (p * q - m * s)
             + w * (m * r - p * p)) / d
        return (a, b, c)


def main():

    # Let's define a small helper function
    def print_me(msg, val):
        print("{}: {}".format(msg, val))

    # Now let's work with the CurveFitting class
    print("\n" + 35 * "*")
    print("*** Use of CurveFitting class")
    print(35 * "*" + "\n")

    # Create a CurveFitting object
    cf1 = CurveFitting(
        [
            73.0,
            38.0,
            35.0,
            42.0,
            78.0,
            68.0,
            74.0,
            42.0,
            52.0,
            54.0,
            39.0,
            61.0,
            42.0,
            49.0,
            50.0,
            62.0,
            44.0,
            39.0,
            43.0,
            54.0,
            44.0,
            37.0,
        ],
        [
            90.4,
            125.3,
            161.8,
            143.4,
            52.5,
            50.8,
            71.5,
            152.8,
            131.3,
            98.5,
            144.8,
            78.1,
            89.5,
            63.9,
            112.1,
            82.0,
            119.8,
            161.2,
            208.4,
            111.6,
            167.1,
            162.1,
        ],
    )

    # Let's use 'linear_fitting()'
    a, b = cf1.linear_fitting()
    print("Linear fitting for cf1:")
    print("   a = {}\tb = {}".format(round(a, 2), round(b, 2)))

    print("")

    # Use the copy constructor
    print("Let's make a copy:")
    cf2 = CurveFitting(cf1)
    print("   cf2 = CurveFitting(cf1)")
    a, b = cf2.linear_fitting()
    print("Linear fitting for cf2:")
    print("   a = {}\tb = {}".format(round(a, 2), round(b, 2)))

    print("")

    # Get the number of value pairs internally stored
    print_me("Number of value pairs inside 'cf2'", len(cf2))  # 22

    print("")

    # Compute the correlation coefficient
    r = cf1.correlation_coeff()
    print("Correlation coefficient:")
    print_me("   r", round(r, 3))

    cf2 = CurveFitting(
        [-2.0, -1.5, -1.0, -0.5, 0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0],
        [
            -9.372,
            -3.821,
            0.291,
            3.730,
            5.822,
            8.324,
            9.083,
            6.957,
            7.006,
            0.365,
            -1.722,
        ],
    )

    print("")

    # Now use 'quadratic_fitting()'
    a, b, c = cf2.quadratic_fitting()
    # Original curve: y = -2.0*x*x + 3.5*x + 7.0 + noise
    print("Quadratic fitting:")
    print("   a = {}\tb = {}\tc = {}".format(round(a, 2), round(b, 2),
                                             round(c, 2)))

    print("")

    cf4 = CurveFitting(
        [
            3,
            20,
            34,
            50,
            75,
            88,
            111,
            129,
            143,
            160,
            183,
            200,
            218,
            230,
            248,
            269,
            290,
            303,
            320,
            344,
        ],
        [
            0.0433,
            0.2532,
            0.3386,
            0.3560,
            0.4983,
            0.7577,
            1.4585,
            1.8628,
            1.8264,
            1.2431,
            -0.2043,
            -1.2431,
            -1.8422,
            -1.8726,
            -1.4889,
            -0.8372,
            -0.4377,
            -0.3640,
            -0.3508,
            -0.2126,
        ],
    )

    # Let's define the three functions to be used for fitting
    def sin1(x):
        return sin(radians(x))

    def sin2(x):
        return sin(radians(2.0 * x))

    def sin3(x):
        return sin(radians(3.0 * x))

    # Use 'general_fitting()' here
    a, b, c = cf4.general_fitting(sin1, sin2, sin3)
    print("General fitting with f0 = sin(x), f1 = sin(2*x), f2 = sin(3*x):")
    print("   a = {}\tb = {}\tc = {}".format(round(a, 2), round(b, 2),
                                             round(c, 2)))

    print("")

    cf5 = CurveFitting([0, 1.2, 1.4, 1.7, 2.1, 2.2])

    a, b, c = cf5.general_fitting(sqrt)
    print("General fitting with f0 = sqrt(x), f1 = 0.0 and f2 = 0.0:")
    print("   a = {}\tb = {}\t\tc = {}".format(round(a, 3), round(b, 3),
                                               round(c, 3)))


if __name__ == "__main__":

    main()
